"""
LightCone class and member functions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2012 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import h5py
import numpy as np
import os

from yt.funcs import *
from yt.analysis_modules.cosmological_observation.cosmology_splice import \
     CosmologySplice
from yt.convenience import \
     load
from yt.utilities.cosmology import \
     Cosmology
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    only_on_root, \
    parallel_objects, \
    parallel_root_only
from yt.visualization.image_writer import \
     write_image
from .common_n_volume import \
     common_volume
from .halo_mask import \
    _light_cone_halo_mask
from .light_cone_projection import \
     _light_cone_projection

class LightCone(CosmologySplice):
    """
    Initialize a LightCone object.

    Parameters
    ----------
    near_redshift : float
        The near (lowest) redshift for the light cone.
    far_redshift : float
        The far (highest) redshift for the light cone.
    observer_redshift : float
        The redshift of the observer.
        Default: 0.0.
    field_of_view_in_arcminutes : float
        The field of view of the image in units of arcminutes.
        Default: 600.0.
    image_resolution_in_arcseconds : float
        The size of each image pixel in units of arcseconds.
        Default: 60.0.
    use_minimum_datasets : bool
        If True, the minimum number of datasets is used to connect the initial
        and final redshift.  If false, the light cone solution will contain
        as many entries as possible within the redshift interval.
        Default: True.
    deltaz_min : float
        Specifies the minimum :math:`\Delta z` between consecutive datasets in
        the returned list.
        Default: 0.0.
    minimum_coherent_box_fraction : float
        Used with use_minimum_datasets set to False, this parameter specifies
        the fraction of the total box size to be traversed before rerandomizing
        the projection axis and center.  This was invented to allow light cones
        with thin slices to sample coherent large scale structure, but in
        practice does not work so well.  Try setting this parameter to 1 and
        see what happens.
        Default: 0.0.
    time_data : bool
        Whether or not to include time outputs when gathering
        datasets for time series.
        Default: True.
    redshift_data : bool
        Whether or not to include redshift outputs when gathering
        datasets for time series.
        Default: True.
    find_outputs : bool
        Whether or not to search for parameter files in the current 
        directory.
        Default: False.
    set_parameters : dict
        Dictionary of parameters to attach to pf.parameters.
        Default: None.
    output_dir : string
        The directory in which images and data files will be written.
        Default: 'LC'.
    output_prefix : string
        The prefix of all images and data files.
        Default: 'LightCone'.

    """
    def __init__(self, parameter_filename, simulation_type,
                 near_redshift, far_redshift,
                 observer_redshift=0.0,
                 field_of_view_in_arcminutes=600.0,
                 image_resolution_in_arcseconds=60.0,
                 use_minimum_datasets=True, deltaz_min=0.0,
                 minimum_coherent_box_fraction=0.0,
                 time_data=True, redshift_data=True,
                 find_outputs=False, set_parameters=None,
                 output_dir='LC', output_prefix='LightCone'):

        self.near_redshift = near_redshift
        self.far_redshift = far_redshift
        self.observer_redshift = observer_redshift
        self.field_of_view_in_arcminutes = field_of_view_in_arcminutes
        self.image_resolution_in_arcseconds = image_resolution_in_arcseconds
        self.use_minimum_datasets = use_minimum_datasets
        self.deltaz_min = deltaz_min
        self.minimum_coherent_box_fraction = minimum_coherent_box_fraction
        if set_parameters is None:
            self.set_parameters = {}
        else:
            self.set_parameters = set_parameters
        self.output_dir = output_dir
        self.output_prefix = output_prefix

        self.master_solution = [] # kept to compare with recycled solutions
        self.projection_stack = []
        self.projection_weight_field_stack = []
        self.halo_mask = []

        # Original random seed of the first solution.
        self.original_random_seed = 0

        # Parameters for recycling light cone solutions.
        self.recycle_solution = False
        self.recycle_random_seed = 0

        # Calculate number of pixels.
        self.pixels = int(self.field_of_view_in_arcminutes * 60.0 / \
                          self.image_resolution_in_arcseconds)

        # Create output directory.
        if not os.path.exists(self.output_dir):
            only_on_root(os.mkdir, self.output_dir)

        # Calculate light cone solution.
        CosmologySplice.__init__(self, parameter_filename, simulation_type,
                                 find_outputs=find_outputs)
        self.light_cone_solution = \
          self.create_cosmology_splice(self.near_redshift, self.far_redshift,
                                       minimal=self.use_minimum_datasets,
                                       deltaz_min=self.deltaz_min,
                                       time_data=time_data,
                                       redshift_data=redshift_data)

    def calculate_light_cone_solution(self, seed=None, filename=None):
        r"""Create list of projections to be added together to make the light cone.

        Several sentences providing an extended description. Refer to
        variables using back-ticks, e.g. `var`.

        Parameters
        ----------
        seed : int
            The seed for the random number generator.  Any light cone solution
            can be reproduced by giving the same random seed.  Default: None
            (each solution will be distinct).
        filename : string
            If given, a text file detailing the solution will be written out.
            Default: None.

        """

        # Don't use box coherence with maximum projection depths.
        if self.use_minimum_datasets and \
                self.minimum_coherent_box_fraction > 0:
            mylog.info("Setting minimum_coherent_box_fraction to 0 with minimal light cone.")
            self.minimum_coherent_box_fraction = 0

        # Make sure recycling flag is off.
        self.recycle_solution = False

        # Get rid of old halo mask, if one was there.
        self.halo_mask = []

        if seed is not None:
            self.original_random_seed = int(seed)

        # Calculate projection sizes, and get
        # random projection axes and centers.
        np.random.seed(self.original_random_seed)

        # For box coherence, keep track of effective depth travelled.
        box_fraction_used = 0.0

        for q in range(len(self.light_cone_solution)):
            if self.light_cone_solution[q].has_key('previous'):
                del self.light_cone_solution[q]['previous']
            if self.light_cone_solution[q].has_key('next'):
                del self.light_cone_solution[q]['next']
            if (q == len(self.light_cone_solution) - 1):
                z_next = self.near_redshift
            else:
                z_next = self.light_cone_solution[q+1]['redshift']

            # Calculate fraction of box required for a depth of delta z
            self.light_cone_solution[q]['box_depth_fraction'] = \
                self.cosmology.ComovingRadialDistance(z_next, \
                        self.light_cone_solution[q]['redshift']) * \
                        self.simulation.hubble_constant / \
                        self.simulation.box_size

            # Simple error check to make sure more than 100% of box depth
            # is never required.
            if (self.light_cone_solution[q]['box_depth_fraction'] > 1.0):
                mylog.debug("Warning: box fraction required to go from z = %f to %f is %f" %
                            (self.light_cone_solution[q]['redshift'], z_next,
                             self.light_cone_solution[q]['box_depth_fraction']))
                mylog.debug("Full box delta z is %f, but it is %f to the next data dump." %
                            (self.light_cone_solution[q]['deltazMax'],
                             self.light_cone_solution[q]['redshift']-z_next))

            # Calculate fraction of box required for width corresponding to
            # requested image size.
            scale = self.cosmology.AngularScale_1arcsec_kpc(self.observer_redshift,
                self.light_cone_solution[q]['redshift'])
            size = self.field_of_view_in_arcminutes * 60.0 * scale / 1000.0
            boxSizeProper = self.simulation.box_size / \
              (self.simulation.hubble_constant *
               (1.0 + self.light_cone_solution[q]['redshift']))
            self.light_cone_solution[q]['box_width_fraction'] = size / boxSizeProper

            # Get projection axis and center.
            # If using box coherence, only get random axis and center if enough
            # of the box has been used, or if box_fraction_used will be greater
            # than 1 after this slice.
            if (q == 0) or (self.minimum_coherent_box_fraction == 0) or \
              (box_fraction_used > self.minimum_coherent_box_fraction) or \
              (box_fraction_used +
               self.light_cone_solution[q]['box_depth_fraction'] > 1.0):
                # Random axis and center.
                self.light_cone_solution[q]['projection_axis'] = \
                  np.random.randint(0, 3)
                self.light_cone_solution[q]['projection_center'] = \
                  [np.random.random() for i in range(3)]
                box_fraction_used = 0.0
            else:
                # Same axis and center as previous slice,
                # but with depth center shifted.
                self.light_cone_solution[q]['projection_axis'] = \
                  self.light_cone_solution[q-1]['projection_axis']
                self.light_cone_solution[q]['projection_center'] = \
                  copy.deepcopy(self.light_cone_solution[q-1]['projection_center'])
                self.light_cone_solution[q]['projection_center']\
                  [self.light_cone_solution[q]['projection_axis']] += \
                    0.5 * (self.light_cone_solution[q]['box_depth_fraction'] +
                           self.light_cone_solution[q-1]['box_depth_fraction'])
                if self.light_cone_solution[q]['projection_center']\
                  [self.light_cone_solution[q]['projection_axis']] >= 1.0:
                    self.light_cone_solution[q]['projection_center']\
                      [self.light_cone_solution[q]['projection_axis']] -= 1.0

            box_fraction_used += self.light_cone_solution[q]['box_depth_fraction']

        # Store this as the master solution.
        self.master_solution = [copy.deepcopy(q) \
                                for q in self.light_cone_solution]

        # Write solution to a file.
        if filename is not None:
            self._save_light_cone_solution(filename=filename)

    def get_halo_mask(self, mask_file=None,
                      cube_file=None, map_file=None,
                      virial_overdensity=200,
                      halo_profiler_parameters=None,
                      njobs=1, dynamic=False):
        r"""Gets a halo mask from a file or makes a new one.

        Parameters
        ----------
        mask_file : string
            An hdf5 file to output the halo mask.
            Default: None
        cub_file : string
            An hdf5 file to output a halo mask for each slice
            of the light cone.
            Default: None
        map_file : string
            A text file to which to output the halo map (locations
            in the images of the halos.
            Default: None
        virial_overdensity : float
            The overdensity at which the virial radius is calculated
            and used as the radial for the halo mask.
            Default: 200
        halo_profiler_parameters: dict
            A dictionary of parameters to be passed to the HaloProfiler
            for each slice of the light cone.
            Default: None
        njobs : int
            The number of parallel jobs over which the slices for the
            halo mask will be split.  Choose -1 for one processor per
            individual slice and 1 to have all processors work together
            on each projection.
            Default: 1
        dynamic : bool
            If True, use dynamic load balancing to create the projections.
            Default: False.

        """

        if halo_profiler_parameters is None:
            halo_profiler_parameters = {}

        # Check if file already exists.
        if mask_file is not None and os.path.exists(mask_file):
            mylog.info('Reading halo mask from %s.' % mask_file)
            in_file = h5py.File(mask_file, 'r')
            self.halo_mask = in_file['HaloMask'].value
            in_file.close()

        # Otherwise, make a halo mask.
        else:
            halo_mask_cube = _light_cone_halo_mask(self, mask_file=mask_file,
                                                   cube_file=cube_file,
                                                   map_file=map_file,
                                                   virial_overdensity=\
                                                   virial_overdensity,
                                                   halo_profiler_parameters=\
                                                   halo_profiler_parameters,
                                                   njobs=njobs,
                                                   dynamic=dynamic)
            # Collapse cube into final mask.
            self.halo_mask = np.ones(shape=(self.pixels, self.pixels),
                                     dtype=bool)
            for mask in halo_mask_cube:
                self.halo_mask *= mask
            del halo_mask_cube

    def project_light_cone(self, field, weight_field=None, apply_halo_mask=False,
                           node=None, save_stack=True, save_final_image=True,
                           save_slice_images=False,
                           cmap_name='algae', photon_field=False,
                           njobs=1, dynamic=False):
        r"""Create projections for light cone, then add them together.

        Parameters
        ----------
        field : string
            The projected field.
        weight_field : string
            the weight field of the projection.  This has the same meaning as
            in standard projections.
            Default: None.
        apply_halo_mask : bool
            if True, a boolean mask is apply to the light cone projection.  See
            below for a description of halo masks.
            Default: False.
        node : string
            a prefix to be prepended to the node name under which the
            projection data is serialized.
            Default: None.
        save_stack : bool
            if True, the light cone data including each individual
            slice is written to an hdf5 file.
            Default: True.
        save_final_image : bool
            if True, save an image of the final light cone projection.
            Default: True.
        save_slice_images : bool
            save images for each individual projection slice.
            Default: False.
        cmap_name : string
            color map for images.
            Default: 'algae'.
        photon_field : bool
            if True, the projection data for each slice is decremented by 4 Pi
            R^2`, where R is the luminosity distance between the observer and
            the slice redshift.
            Default: False.
        njobs : int
            The number of parallel jobs over which the light cone projection
            will be split.  Choose -1 for one processor per individual
            projection and 1 to have all processors work together on each
            projection.
            Default: 1.
        dynamic : bool
            If True, use dynamic load balancing to create the projections.
            Default: False.

        """

        # Clear projection stack.
        self.projection_stack = []
        self.projection_weight_field_stack = []
        if (self.light_cone_solution[-1].has_key('object')):
            del self.light_cone_solution[-1]['object']

        if not(self.output_dir.endswith("/")):
                 self.output_dir += "/"

        # for q, output in enumerate(self.light_cone_solution):
        all_storage = {}
        for my_storage, output in parallel_objects(self.light_cone_solution,
                                                   storage=all_storage,
                                                   dynamic=dynamic):
            output['object'] = load(output['filename'])
            output['object'].parameters.update(self.set_parameters)
            frb = _light_cone_projection(output, field, self.pixels,
                                         weight_field=weight_field, node=node)

            if photon_field:
                # Decrement the flux by the luminosity distance.
                # Assume field in frb is in erg/s/cm^2/Hz
                dL = self.cosmology.LuminosityDistance(self.observer_redshift,
                                                       output['redshift']) #in Mpc
                boxSizeProper = self.simulation.box_size / \
                  (self.simulation.hubble_constant * (1.0 + output['redshift']))
                pixelarea = (boxSizeProper/self.pixels)**2 #in proper cm^2
                factor = pixelarea/(4.0*np.pi*dL**2)
                mylog.info("Distance to slice = %e" % dL)
                frb[field] *= factor #in erg/s/cm^2/Hz on observer's image plane.

            if weight_field is None:
                my_storage.result = {'field': frb[field]}
            else:
                my_storage.result = {'field': (frb[field] *
                                               frb['weight_field']),
                                     'weight_field': frb['weight_field']}

            del output['object']

        # Combine results from each slice.
        all_slices = all_storage.keys()
        all_slices.sort()
        for my_slice in all_slices:
            if save_slice_images:
                if node is None:
                    name = "%s%s_%04d_%04d" % (self.output_dir,
                                               self.output_prefix,
                                               my_slice,
                                               len(self.light_cone_solution))
                else:
                    name = "%s%s_%s_%04d_%04d" % (self.output_dir,
                                                  self.output_prefix,
                                                  node, my_slice,
                                                  len(self.light_cone_solution))
                if weight_field is None:
                    my_image = all_storage[my_slice]['field']
                else:
                    my_image = all_storage[my_slice]['field'] / \
                      all_storage[my_slice]['weight_field']
                only_on_root(write_image, np.log10(my_image),
                             "%s_%s.png" % (name, field), cmap_name=cmap_name)

            self.projection_stack.append(all_storage[my_slice]['field'])
            if weight_field is not None:
                self.projection_weight_field_stack.append(all_storage[my_slice]['field'])

        # Add up slices to make light cone projection.
        if (weight_field is None):
            light_cone_projection = sum(self.projection_stack)
        else:
            light_cone_projection = sum(self.projection_stack) / \
              sum(self.projection_weight_field_stack)

        if node is None:
            filename = "%s%s" % (self.output_dir, self.output_prefix)
        else:
            filename = "%s%s_%s" % (self.output_dir, self.output_prefix, node)

        # Apply halo mask.
        if apply_halo_mask:
            if len(self.halo_mask) > 0:
                mylog.info("Applying halo mask.")
                light_cone_projection *= self.halo_mask
            else:
                mylog.error("No halo mask loaded, call get_halo_mask.")


        # Write image.
        if save_final_image:
            only_on_root(write_image, np.log10(light_cone_projection),
                         "%s_%s.png" % (filename, field), cmap_name=cmap_name)

        # Write stack to hdf5 file.
        if save_stack:
            self._save_light_cone_stack(field=field, weight_field=weight_field,
                                        filename=filename)

    def rerandomize_light_cone_solution(self, new_seed, recycle=True, filename=None):
        """
        When making a projection for a light cone, only randomizations along the
        line of sight make any given projection unique, since the lateral shifting
        and tiling is done after the projection is made.  Therefore, multiple light
        cones can be made from a single set of projections by introducing different
        lateral random shifts and keeping all the original shifts along the line of
        sight.  This routine will take in a new random seed and rerandomize the
        parts of the light cone that do not contribute to creating a unique
        projection object.  Additionally, this routine is built such that if the
        same random seed is given for the rerandomizing, the solution will be
        identical to the original.

        This routine has now been updated to be a general solution rescrambler.
        If the keyword recycle is set to True, then it will recycle.  Otherwise, it
        will create a completely new solution.

        new_sed : float
            The new random seed.
        recycle : bool
            If True, the new solution will have the same shift in the line of
            sight as the original solution.  Since the projections of each
            slice are serialized and stored for the entire width of the box
            (even if the width used is left than the total box), the projection
            data can be deserialized instead of being remade from scratch.
            This can greatly speed up the creation of a large number of light
            cone projections.
            Default: True.
        filename : string
            If given, a text file detailing the solution will be written out.
            Default: None.

        """

        # Get rid of old halo mask, if one was there.
        self.halo_mask = []

        # Clean pf objects out of light cone solution.
        for my_slice in self.light_cone_solution:
            if my_slice.has_key('object'):
                del my_slice['object']

        if recycle:
            mylog.debug("Recycling solution made with %s with new seed %s." %
                        (self.original_random_seed, new_seed))
            self.recycle_random_seed = int(new_seed)
        else:
            mylog.debug("Creating new solution with random seed %s." % new_seed)
            self.original_random_seed = int(new_seed)
            self.recycle_random_seed = 0

        self.recycle_solution = recycle

        # Keep track of fraction of volume in common between the original and
        # recycled solution.
        my_volume = 0.0
        total_volume = 0.0

        # For box coherence, keep track of effective depth travelled.
        box_fraction_used = 0.0

        # Seed random number generator with new seed.
        np.random.seed(int(new_seed))

        for q, output in enumerate(self.light_cone_solution):
            # It is necessary to make the same number of calls to the random
            # number generator so the original solution willbe produced if the
            # same seed is given.

            # Get projection axis and center.
            # If using box coherence, only get random axis and center if enough
            # of the box has been used, or if box_fraction_used will be greater
            # than 1 after this slice.
            if (q == 0) or (self.minimum_coherent_box_fraction == 0) or \
                    (box_fraction_used > self.minimum_coherent_box_fraction) or \
                    (box_fraction_used + self.light_cone_solution[q]['box_depth_fraction'] > 1.0):
                # Get random projection axis and center.
                # If recycling, axis will get thrown away since it is used in
                # creating a unique projection object.
                newAxis = np.random.randint(0, 3)

                newCenter = [np.random.random() for i in range(3)]
                box_fraction_used = 0.0
            else:
                # Same axis and center as previous slice, but with depth center shifted.
                newAxis = self.light_cone_solution[q-1]['projection_axis']
                newCenter = copy.deepcopy(self.light_cone_solution[q-1]['projection_center'])
                newCenter[newAxis] += \
                    0.5 * (self.light_cone_solution[q]['box_depth_fraction'] +
                           self.light_cone_solution[q-1]['box_depth_fraction'])
                if newCenter[newAxis] >= 1.0:
                    newCenter[newAxis] -= 1.0

            if recycle:
                output['projection_axis'] = self.master_solution[q]['projection_axis']
            else:
                output['projection_axis'] = newAxis

            box_fraction_used += self.light_cone_solution[q]['box_depth_fraction']

            # Make list of rectangle corners to calculate common volume.
            newCube = np.zeros(shape=(len(newCenter), 2))
            oldCube = np.zeros(shape=(len(newCenter), 2))
            for w in range(len(newCenter)):
                if (w == self.master_solution[q]['projection_axis']):
                    oldCube[w] = [self.master_solution[q]['projection_center'][w] -
                                  0.5 * self.master_solution[q]['box_depth_fraction'],
                                  self.master_solution[q]['projection_center'][w] +
                                  0.5 * self.master_solution[q]['box_depth_fraction']]
                else:
                    oldCube[w] = [self.master_solution[q]['projection_center'][w] -
                                  0.5 * self.master_solution[q]['box_width_fraction'],
                                  self.master_solution[q]['projection_center'][w] +
                                  0.5 * self.master_solution[q]['box_width_fraction']]

                if (w == output['projection_axis']):
                    if recycle:
                        newCube[w] = oldCube[w]
                    else:
                        newCube[w] = \
                          [newCenter[w] -
                           0.5 * self.master_solution[q]['box_depth_fraction'],
                           newCenter[w] +
                           0.5 * self.master_solution[q]['box_depth_fraction']]
                else:
                    newCube[w] = [newCenter[w] -
                                  0.5 * self.master_solution[q]['box_width_fraction'],
                                  newCenter[w] +
                                  0.5 * self.master_solution[q]['box_width_fraction']]

            my_volume += common_volume(oldCube, newCube,
                                           periodic=np.array([[0, 1],
                                                              [0, 1],
                                                              [0, 1]]))
            total_volume += output['box_depth_fraction'] * \
              output['box_width_fraction']**2

            # Replace centers for every axis except the line of sight axis.
            for w in range(len(newCenter)):
                if not(recycle and
                       (w == self.light_cone_solution[q]['projection_axis'])):
                    self.light_cone_solution[q]['projection_center'][w] = \
                      newCenter[w]

        if recycle:
            mylog.debug("Fractional common volume between master and recycled solution is %.2e" % \
                        (my_volume / total_volume))
        else:
            mylog.debug("Fraction of total volume in common with old solution is %.2e." % \
                        (my_volume / total_volume))
            self.master_solution = [copy.deepcopy(q) \
                                    for q in self.light_cone_solution]

        # Write solution to a file.
        if filename is not None:
            self._save_light_cone_solution(filename=filename)

    def restore_master_solution(self):
        "Reset the active light cone solution to the master solution."
        self.light_cone_solution = [copy.deepcopy(q) \
                                    for q in self.master_solution]

    @parallel_root_only
    def _save_light_cone_solution(self, filename="light_cone.dat"):
        "Write out a text file with information on light cone solution."

        mylog.info("Saving light cone solution to %s." % filename)

        f = open(filename, 'w')
        if self.recycle_solution:
            f.write("Recycled Solution\n")
            f.write("OriginalRandomSeed = %s\n" % self.original_random_seed)
            f.write("RecycleRandomSeed = %s\n" % self.recycle_random_seed)
        else:
            f.write("Original Solution\n")
            f.write("OriginalRandomSeed = %s\n" % self.original_random_seed)
        f.write("parameter_filename = %s\n" % self.parameter_filename)
        f.write("\n")
        for q, output in enumerate(self.light_cone_solution):
            f.write("Proj %04d, %s, z = %f, depth/box = %f, width/box = %f, axis = %d, center = %f, %f, %f\n" %
                    (q, output['filename'], output['redshift'],
                     output['box_depth_fraction'], output['box_width_fraction'],
                     output['projection_axis'], output['projection_center'][0],
                     output['projection_center'][1], output['projection_center'][2]))
        f.close()

    @parallel_root_only
    def _save_light_cone_stack(self, field=None, weight_field=None,
                               filename=None, over_write=True):
        "Save the light cone projection stack as a 3d array in and hdf5 file."

        # Make list of redshifts to include as a dataset attribute.
        redshiftList = np.array([my_slice['redshift'] \
                                 for my_slice in self.light_cone_solution])

        field_node = "%s_%s" % (field, weight_field)
        weight_field_node = "weight_field_%s" % weight_field

        if (filename is None):
            filename = "%s/%s_data" % (self.output_dir, self.output_prefix)
        if not(filename.endswith('.h5')):
               filename += ".h5"

        if (len(self.projection_stack) == 0):
            mylog.debug("save_light_cone_stack: no projection data loaded.")
            return

        mylog.info("Writing light cone data to %s." % filename)

        output = h5py.File(filename, "a")

        node_exists = field_node in output

        if node_exists:
            if over_write:
                mylog.info("Dataset, %s, already exists, overwriting." %
                           field_node)
                write_data = True
                del output[field_node]
            else:
                mylog.info("Dataset, %s, already exists in %s, not saving." %
                           (field_node, filename))
                write_data = False
        else:
            write_data = True

        if write_data:
            mylog.info("Saving %s to %s." % (field_node, filename))
            self.projection_stack = np.array(self.projection_stack)
            field_dataset = output.create_dataset(field_node,
                                                  data=self.projection_stack)
            field_dataset.attrs['redshifts'] = redshiftList
            field_dataset.attrs['observer_redshift'] = \
              np.float(self.observer_redshift)
            field_dataset.attrs['field_of_view_in_arcminutes'] = \
              np.float(self.field_of_view_in_arcminutes)
            field_dataset.attrs['image_resolution_in_arcseconds'] = \
              np.float(self.image_resolution_in_arcseconds)

        if (len(self.projection_weight_field_stack) > 0):
            if node_exists:
                if over_write:
                    mylog.info("Dataset, %s, already exists, overwriting." %
                               weight_field_node)
                    del output[field_node]
                else:
                    mylog.info("Dataset, %s, already exists in %s, not saving." %
                               (weight_field_node, filename))
                    write_data = False
            else:
                write_data = True

            if write_data:
                mylog.info("Saving %s to %s." % (weight_field_node, filename))
                self.projection_weight_field_stack = \
                  np.array(self.projection_weight_field_stack)
                weight_field_dataset = \
                  output.create_dataset(weight_field_node,
                                        data=self.projection_weight_field_stack)
                weight_field_dataset.attrs['redshifts'] = redshiftList
                weight_field_dataset.attrs['observer_redshift'] = \
                  np.float(self.observer_redshift)
                weight_field_dataset.attrs['field_of_view_in_arcminutes'] = \
                  np.float(self.field_of_view_in_arcminutes)
                weight_field_dataset.attrs['image_resolution_in_arcseconds'] = \
                  np.float(self.image_resolution_in_arcseconds)

        output.close()



import os
import os.path
import glob
import sys
import time
import subprocess
import shutil
import glob
import distribute_setup
distribute_setup.use_setuptools()

from distutils.command.build_py import build_py
from numpy.distutils.misc_util import appendpath
from numpy.distutils.command import install_data as np_install_data
from numpy.distutils import log
from distutils import version

from distutils.core import Command
from distutils.spawn import find_executable

def find_fortran_deps():
    return (find_executable("Forthon"),
            find_executable("gfortran"))

class BuildForthon(Command):

    """Command for building Forthon modules"""

    description = "Build Forthon modules"
    user_options = []

    def initialize_options(self):

        """init options"""

        pass

    def finalize_options(self):

        """finalize options"""

        pass

    def run(self):

        """runner"""
        (Forthon_exe, gfortran_exe) = find_fortran_deps()
        if None in (Forthon_exe, gfortran_exe):
            sys.stderr.write(
                "fKDpy.so won't be built due to missing Forthon/gfortran\n"
            )
            return

        cwd = os.getcwd()
        os.chdir(os.path.join(cwd, 'yt/utilities/kdtree'))
        cmd = [Forthon_exe, "-F", "gfortran", "--compile_first",
               "fKD_source", "--no2underscores", "--fopt", "'-O3'", "fKD",
               "fKD_source.f90"]
        subprocess.check_call(cmd, shell=False)
        shutil.move(glob.glob('build/lib*/fKDpy.so')[0], os.getcwd())
        os.chdir(cwd)

REASON_FILES = []
REASON_DIRS = [
    "",
    "resources",
    "resources/ux",
    "resources/images",
    "resources/css",
    "resources/css/images",
    "app",
    "app/store",
    "app/store/widgets",
    "app/view",
    "app/view/widgets",
    "app/model",
    "app/controller",
    "app/controller/widgets",
    "app/templates",
]

for subdir in REASON_DIRS:
    dir_name = "yt/gui/reason/html/%s/" % (subdir)
    files = []
    for ext in ["js", "html", "css", "png", "ico", "gif"]:
        files += glob.glob("%s/*.%s" % (dir_name, ext))
    REASON_FILES.append((dir_name, files))

# Verify that we have Cython installed
try:
    import Cython
    if version.LooseVersion(Cython.__version__) < version.LooseVersion('0.16'):
        needs_cython = True
    else:
        needs_cython = False
except ImportError as e:
    needs_cython = True

if needs_cython:
    print "Cython is a build-time requirement for the source tree of yt."
    print "Please either install yt from a provided, release tarball,"
    print "or install Cython (version 0.16 or higher)."
    print "You may be able to accomplish this by typing:"
    print "     pip install -U Cython"
    sys.exit(1)

######
# This next bit comes from Matthew Brett, to get Cython working with NumPy
# distutils.  I added a bit to get C++ Cython working.
from os.path import join as pjoin, dirname
from distutils.dep_util import newer_group
from distutils.errors import DistutilsError


def generate_a_pyrex_source(self, base, ext_name, source, extension):
    ''' Monkey patch for numpy build_src.build_src method

    Uses Cython instead of Pyrex.

    Assumes Cython is present
    '''
    if self.inplace:
        target_dir = dirname(base)
    else:
        target_dir = appendpath(self.build_src, dirname(base))
    if extension.language == "c++":
        cplus = True
        file_ext = ".cpp"
    else:
        cplus = False
        file_ext = ".c"
    target_file = pjoin(target_dir, ext_name + file_ext)
    depends = [source] + extension.depends
    if self.force or newer_group(depends, target_file, 'newer'):
        import Cython.Compiler.Main
        log.info("cythonc:> %s" % (target_file))
        self.mkpath(target_dir)
        options = Cython.Compiler.Main.CompilationOptions(
            defaults=Cython.Compiler.Main.default_options,
            include_path=extension.include_dirs,
            language=extension.language, cplus=cplus,
            output_file=target_file)
        cython_result = Cython.Compiler.Main.compile(source,
                                                     options=options)
        if cython_result.num_errors != 0:
            raise DistutilsError("%d errors while compiling %r with Cython"
                                 % (cython_result.num_errors, source))
    return target_file


from numpy.distutils.command import build_src
build_src.build_src.generate_a_pyrex_source = generate_a_pyrex_source
# End snippet
######

import setuptools

VERSION = "2.5.4"

if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')


def get_mercurial_changeset_id(target_dir):
    """adapted from a script by Jason F. Harris, published at

    http://jasonfharris.com/blog/2010/05/versioning-your-application-with-the-mercurial-changeset-hash/

    """
    import subprocess
    import re
    get_changeset = subprocess.Popen('hg identify -b -i',
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     shell=True)

    if (get_changeset.stderr.read() != ""):
        print "Error in obtaining current changeset of the Mercurial repository"
        changeset = None

    changeset = get_changeset.stdout.read().strip()
    if (not re.search("^[0-9a-f]{12}", changeset)):
        print "Current changeset of the Mercurial repository is malformed"
        changeset = None

    return changeset


class my_build_src(build_src.build_src):
    def run(self):
        self.run_command("build_forthon")
        build_src.build_src.run(self)


class my_install_data(np_install_data.install_data):
    def run(self):
        (Forthon_exe, gfortran_exe) = find_fortran_deps()
        if None in (Forthon_exe, gfortran_exe):
            pass
        else:
            self.distribution.data_files.append(
                ('yt/utilities/kdtree', ['yt/utilities/kdtree/fKDpy.so'])
                )
        np_install_data.install_data.run(self)

class my_build_py(build_py):
    def run(self):
        # honor the --dry-run flag
        if not self.dry_run:
            target_dir = os.path.join(self.build_lib, 'yt')
            src_dir = os.getcwd()
            changeset = get_mercurial_changeset_id(src_dir)
            self.mkpath(target_dir)
            with open(os.path.join(target_dir, '__hg_version__.py'), 'w') as fobj:
                fobj.write("hg_version = '%s'\n" % changeset)

            build_py.run(self)


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.make_config_py()
    # config.make_svn_version_py()
    config.add_subpackage('yt', 'yt')
    config.add_scripts("scripts/*")

    return config


def setup_package():

    from numpy.distutils.core import setup

    setup(
        name="yt",
        version=VERSION,
        description="An analysis and visualization toolkit for Astrophysical "
                    + "simulations.",
        classifiers=["Development Status :: 5 - Production/Stable",
                     "Environment :: Console",
                     "Intended Audience :: Science/Research",
                     "License :: OSI Approved :: GNU General Public License (GPL)",
                     "Operating System :: MacOS :: MacOS X",
                     "Operating System :: POSIX :: AIX",
                     "Operating System :: POSIX :: Linux",
                     "Programming Language :: C",
                     "Programming Language :: Python",
                     "Topic :: Scientific/Engineering :: Astronomy",
                     "Topic :: Scientific/Engineering :: Physics",
                     "Topic :: Scientific/Engineering :: Visualization"],
        keywords='astronomy astrophysics visualization ' +
        'amr adaptivemeshrefinement',
        entry_points={'console_scripts': [
        'yt = yt.utilities.command_line:run_main',
        ],
            'nose.plugins.0.10': [
                'answer-testing = yt.utilities.answer_testing.framework:AnswerTesting'
            ]
        },
        author="Matthew J. Turk",
        author_email="matthewturk@gmail.com",
        url="http://yt-project.org/",
        license="GPL-3",
        configuration=configuration,
        zip_safe=False,
        data_files=REASON_FILES,
        cmdclass={'build_py': my_build_py, 'build_forthon': BuildForthon,
                  'build_src': my_build_src, 'install_data': my_install_data},
    )
    return

if __name__ == '__main__':
    setup_package()



"""
Data structures for GDF.

Author: Samuel W. Skillman <samskillman@gmail.com>
Affiliation: University of Colorado at Boulder
Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Samuel W. Skillman, Matthew Turk, J. S. Oishi.
  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import numpy as np
import weakref
from yt.funcs import *
from yt.data_objects.grid_patch import \
           AMRGridPatch
from yt.data_objects.hierarchy import \
           AMRHierarchy
from yt.data_objects.static_output import \
           StaticOutput
from yt.utilities.lib import \
    get_box_grids_level
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion

from .fields import GDFFieldInfo, KnownGDFFields
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
import pdb

def _get_convert(fname):
    def _conv(data):
        return data.convert(fname)
    return _conv

class GDFGrid(AMRGridPatch):
    _id_offset = 0
    def __init__(self, id, hierarchy, level, start, dimensions):
        AMRGridPatch.__init__(self, id, filename = hierarchy.hierarchy_filename,
                              hierarchy = hierarchy)
        self.Parent = []
        self.Children = []
        self.Level = level
        self.start_index = start.copy()
        self.stop_index = self.start_index + dimensions
        self.ActiveDimensions = dimensions.copy()

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        if len(self.Parent) > 0:
            self.dds = self.Parent[0].dds / self.pf.refine_by
        else:
            LE, RE = self.hierarchy.grid_left_edge[id,:], \
                     self.hierarchy.grid_right_edge[id,:]
            self.dds = np.array((RE-LE)/self.ActiveDimensions)
        if self.pf.data_software != "piernik":
            if self.pf.dimensionality < 2: self.dds[1] = 1.0
            if self.pf.dimensionality < 3: self.dds[2] = 1.0
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

class GDFHierarchy(AMRHierarchy):

    grid = GDFGrid

    def __init__(self, pf, data_style='grid_data_format'):
        self.parameter_file = weakref.proxy(pf)
        self.data_style = data_style
        # for now, the hierarchy file is the parameter file!
        self.hierarchy_filename = self.parameter_file.parameter_filename
        self.directory = os.path.dirname(self.hierarchy_filename)
        self._fhandle = h5py.File(self.hierarchy_filename,'r')
        AMRHierarchy.__init__(self,pf,data_style)

        self._fhandle.close()

    def _initialize_data_storage(self):
        pass

    def _detect_fields(self):
        self.field_list = self._fhandle['field_types'].keys()

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        self.num_grids = self._fhandle['/grid_parent_id'].shape[0]

    def _parse_hierarchy(self):
        f = self._fhandle
        dxs = []
        self.grids = np.empty(self.num_grids, dtype='object')
        levels = (f['grid_level'][:]).copy()
        glis = (f['grid_left_index'][:]).copy()
        gdims = (f['grid_dimensions'][:]).copy()
        active_dims = ~((np.max(gdims, axis=0) == 1) &
                        (self.parameter_file.domain_dimensions == 1))

        for i in range(levels.shape[0]):
            self.grids[i] = self.grid(i, self, levels[i],
                                      glis[i],
                                      gdims[i])
            self.grids[i]._level_id = levels[i]

            dx = (self.parameter_file.domain_right_edge-
                  self.parameter_file.domain_left_edge)/self.parameter_file.domain_dimensions
            dx[active_dims] = dx[active_dims]/self.parameter_file.refine_by**(levels[i])
            dxs.append(dx)
        dx = np.array(dxs)
        self.grid_left_edge = self.parameter_file.domain_left_edge + dx*glis
        self.grid_dimensions = gdims.astype("int32")
        self.grid_right_edge = self.grid_left_edge + dx*self.grid_dimensions
        self.grid_particle_count = f['grid_particle_count'][:]
        del levels, glis, gdims

    def _populate_grid_objects(self):
        mask = np.empty(self.grids.size, dtype='int32')
        for gi, g in enumerate(self.grids):
            g._prepare_grid()
            g._setup_dx()

        for gi, g in enumerate(self.grids):
            g.Children = self._get_grid_children(g)
            for g1 in g.Children:
                g1.Parent.append(g)
            get_box_grids_level(self.grid_left_edge[gi,:],
                                self.grid_right_edge[gi,:],
                                self.grid_levels[gi],
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask)
            m = mask.astype("bool")
            m[gi] = False
            siblings = self.grids[gi:][m[gi:]]
            if len(siblings) > 0:
                g.OverlappingSiblings = siblings.tolist()
        self.max_level = self.grid_levels.max()

    def _setup_derived_fields(self):
        self.derived_field_list = []

    def _get_grid_children(self, grid):
        mask = np.zeros(self.num_grids, dtype='bool')
        grids, grid_ind = self.get_box_grids(grid.LeftEdge, grid.RightEdge)
        mask[grid_ind] = True
        return [g for g in self.grids[mask] if g.Level == grid.Level + 1]

class GDFStaticOutput(StaticOutput):
    _hierarchy_class = GDFHierarchy
    _fieldinfo_fallback = GDFFieldInfo
    _fieldinfo_known = KnownGDFFields

    def __init__(self, filename, data_style='grid_data_format',
                 storage_filename = None):
        StaticOutput.__init__(self, filename, data_style)
        self.storage_filename = storage_filename
        self.filename = filename

    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['cm'] = 1.0
        self.units['unitary'] = 1.0 / (self.domain_right_edge - self.domain_left_edge).max()
        for unit in mpc_conversion.keys():
            self.units[unit] = 1.0 * mpc_conversion[unit] / mpc_conversion["cm"]
        for unit in sec_conversion.keys():
            self.time_units[unit] = 1.0 / sec_conversion[unit]

        # This should be improved.
        self._handle = h5py.File(self.parameter_filename, "r")
        for field_name in self._handle["/field_types"]:
            current_field = self._handle["/field_types/%s" % field_name]
            if 'field_to_cgs' in current_field.attrs:
                self.units[field_name] = current_field.attrs['field_to_cgs']
            else:
                self.units[field_name] = 1.0
            if 'field_units' in current_field.attrs:
                current_fields_unit = just_one(current_field.attrs['field_units'])
            else:
                current_fields_unit = ""
            self._fieldinfo_known.add_field(field_name, function=NullFunc, take_log=False,
                   units=current_fields_unit, projected_units="",
                   convert_function=_get_convert(field_name))

        self._handle.close()
        del self._handle

    def _parse_parameter_file(self):
        self._handle = h5py.File(self.parameter_filename, "r")
        if 'data_software' in self._handle['gridded_data_format'].attrs:
            self.data_software = \
                self._handle['gridded_data_format'].attrs['data_software']
        else:
            self.data_software = "unknown"
        sp = self._handle["/simulation_parameters"].attrs
        self.domain_left_edge = sp["domain_left_edge"][:]
        self.domain_right_edge = sp["domain_right_edge"][:]
        self.domain_dimensions = sp["domain_dimensions"][:]
        refine_by = sp["refine_by"]
        if refine_by is None: refine_by = 2
        self.refine_by = refine_by
        self.dimensionality = sp["dimensionality"]
        self.current_time = sp["current_time"]
        self.unique_identifier = sp["unique_identifier"]
        self.cosmological_simulation = sp["cosmological_simulation"]
        if sp["num_ghost_zones"] != 0: raise RuntimeError
        self.num_ghost_zones = sp["num_ghost_zones"]
        self.field_ordering = sp["field_ordering"]
        self.boundary_conditions = sp["boundary_conditions"][:]
        p = [bnd == 0 for bnd in self.boundary_conditions[::2]]
        self.periodicity = ensure_tuple(p)
        if self.cosmological_simulation:
            self.current_redshift = sp["current_redshift"]
            self.omega_lambda = sp["omega_lambda"]
            self.omega_matter = sp["omega_matter"]
            self.hubble_constant = sp["hubble_constant"]
        else:
            self.current_redshift = self.omega_lambda = self.omega_matter = \
                self.hubble_constant = self.cosmological_simulation = 0.0
        self.parameters['Time'] = 1.0 # Hardcode time conversion for now.
        self.parameters["HydroMethod"] = 0 # Hardcode for now until field staggering is supported.
        self._handle.close()
        del self._handle

    @classmethod
    def _is_valid(self, *args, **kwargs):
        try:
            fileh = h5py.File(args[0],'r')
            if "gridded_data_format" in fileh:
                fileh.close()
                return True
            fileh.close()
        except:
            pass
        return False

    def __repr__(self):
        return self.basename.rsplit(".", 1)[0]




"""
Profile classes, to deal with generating and obtaining profiles

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Samuel Skillman <samskillman@gmail.com>
Affiliation: CASA, University of Colorado at Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import numpy as np

from yt.funcs import *

from yt.data_objects.data_containers import YTFieldData
from yt.utilities.lib import bin_profile1d, bin_profile2d, bin_profile3d
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface

_field_mapping = {
    "total_mass": ("CellMassMsun", "ParticleMassMsun"),
    "hybrid_radius": ("RadiusCode", "ParticleRadiusCode"),
                 }

class EmptyProfileData(Exception):
    pass

def preserve_source_parameters(func):
    def save_state(*args, **kwargs):
        # Temporarily replace the 'field_parameters' for a
        # grid with the 'field_parameters' for the data source
        prof = args[0]
        source = args[1]
        if hasattr(source, 'field_parameters'):
            old_params = source.field_parameters
            source.field_parameters = prof._data_source.field_parameters
            tr = func(*args, **kwargs)
            source.field_parameters = old_params
        else:
            tr = func(*args, **kwargs)
        return tr
    return save_state

# Note we do not inherit from EnzoData.
# We could, but I think we instead want to deal with the root datasource.
class BinnedProfile(ParallelAnalysisInterface):
    def __init__(self, data_source, lazy_reader):
        ParallelAnalysisInterface.__init__(self)
        self._data_source = data_source
        self.pf = data_source.pf
        self.field_data = YTFieldData()
        self._pdata = {}
        self._lazy_reader = lazy_reader

    @property
    def hierarchy(self):
        return self.pf.hierarchy

    def _get_dependencies(self, fields):
        return ParallelAnalysisInterface._get_dependencies(
                    self, fields + self._get_bin_fields())

    def _initialize_parallel(self, fields):
        g_objs = [g for g in self._get_grid_objs()]
        self.comm.preload(g_objs, self.get_dependencies(fields),
                      self._data_source.hierarchy.io)

    def _lazy_add_fields(self, fields, weight, accumulation):
        self._ngrids = 0
        self.__data = {}         # final results will go here
        self.__weight_data = {}  # we need to track the weights as we go
        self.__std_data = {}
        for field in fields:
            self.__data[field] = self._get_empty_field()
            self.__weight_data[field] = self._get_empty_field()
            self.__std_data[field] = self._get_empty_field()
        self.__used = self._get_empty_field().astype('bool')
        #pbar = get_pbar('Binning grids', len(self._data_source._grids))
        for gi,grid in enumerate(self._get_grids(fields)):
            self._ngrids += 1
            #pbar.update(gi)
            try:
                args = self._get_bins(grid, check_cut=True)
            except EmptyProfileData:
                # No bins returned for this grid, so forget it!
                continue
            for field in fields:
                # We get back field values, weight values, used bins
                f, w, q, u = self._bin_field(grid, field, weight, accumulation,
                                          args=args, check_cut=True)
                self.__data[field] += f        # running total
                self.__weight_data[field] += w # running total
                self.__std_data[field][u] += w[u] * (q[u]/w[u] + \
                    (f[u]/w[u] -
                     self.__data[field][u]/self.__weight_data[field][u])**2) # running total
                self.__used = (self.__used | u)       # running 'or'
            grid.clear_data()
        # When the loop completes the parallel finalizer gets called
        #pbar.finish()
        ub = np.where(self.__used)
        for field in fields:
            if weight: # Now, at the end, we divide out.
                self.__data[field][ub] /= self.__weight_data[field][ub]
                self.__std_data[field][ub] /= self.__weight_data[field][ub]
            self[field] = self.__data[field]
            self["%s_std" % field] = np.sqrt(self.__std_data[field])
        self["UsedBins"] = self.__used
        del self.__data, self.__std_data, self.__weight_data, self.__used

    def _finalize_parallel(self):
        my_mean = {}
        my_weight = {}
        for key in self.__data:
            my_mean[key] = self._get_empty_field()
            my_weight[key] = self._get_empty_field()
        ub = np.where(self.__used)
        for key in self.__data:
            my_mean[key][ub] = self.__data[key][ub] / self.__weight_data[key][ub]
            my_weight[key][ub] = self.__weight_data[key][ub]
        for key in self.__data:
            self.__data[key] = self.comm.mpi_allreduce(self.__data[key], op='sum')
        for key in self.__weight_data:
            self.__weight_data[key] = self.comm.mpi_allreduce(self.__weight_data[key], op='sum')
        for key in self.__std_data:
            self.__std_data[key][ub] = my_weight[key][ub] * (self.__std_data[key][ub] / my_weight[key][ub] + \
                (my_mean[key][ub] - self.__data[key][ub]/self.__weight_data[key][ub])**2)
            self.__std_data[key] = self.comm.mpi_allreduce(self.__std_data[key], op='sum')
        self.__used = self.comm.mpi_allreduce(self.__used, op='sum')

    def _unlazy_add_fields(self, fields, weight, accumulation):
        for field in fields:
            f, w, q, u = self._bin_field(self._data_source, field, weight,
                                         accumulation, self._args, check_cut = False)
            if weight:
                f[u] /= w[u]
                q[u] = np.sqrt(q[u] / w[u])
            self[field] = f
            self["%s_std" % field] = q
        self["UsedBins"] = u

    def add_fields(self, fields, weight = "CellMassMsun", accumulation = False, fractional=False):
        """
        We accept a list of *fields* which will be binned if *weight* is not
        None and otherwise summed.  *accumulation* determines whether or not
        they will be accumulated from low to high along the appropriate axes.
        """
        # Note that the specification has to be the same for all of these
        fields = ensure_list(fields)
        if self._lazy_reader:
            self._lazy_add_fields(fields, weight, accumulation)
        else:
            self._unlazy_add_fields(fields, weight, accumulation)
        if fractional:
            for field in fields:
                self.field_data[field] /= self.field_data[field].sum()

    def keys(self):
        return self.field_data.keys()

    def __getitem__(self, key):
        # This raises a KeyError if it doesn't exist
        # This is because we explicitly want to add all fields
        return self.field_data[key]

    def __setitem__(self, key, value):
        self.field_data[key] = value

    def _get_field(self, source, this_field, check_cut):
        # This is where we will iterate to get all contributions to a field
        # which is how we will implement hybrid particle/cell fields
        # but...  we default to just the field.
        data = []
        for field in _field_mapping.get(this_field, (this_field,)):
            pointI = None
            if check_cut:
                # This conditional is so that we can have variable-length
                # particle fields.  Note that we can't apply the
                # is_fully_enclosed to baryon fields, because child cells get
                # in the way.
                if field in self.pf.field_info \
                    and self.pf.field_info[field].particle_type:
                    if not self._data_source._is_fully_enclosed(source):
                        pointI = self._data_source._get_particle_indices(source)
                else:
                    pointI = self._data_source._get_point_indices(source)
            data.append(source[field][pointI].ravel().astype('float64'))
        return np.concatenate(data, axis=0)

    def _fix_pickle(self):
        if isinstance(self._data_source, tuple):
            self._data_source = self._data_source[1]

# @todo: Fix accumulation with overriding
class BinnedProfile1D(BinnedProfile):
    """
    A 'Profile' produces either a weighted (or unweighted) average or a
    straight sum of a field in a bin defined by another field.  In the case
    of a weighted average, we have: p_i = sum( w_i * v_i ) / sum(w_i)

    We accept a *data_source*, which will be binned into *n_bins*
    by the field *bin_field* between the *lower_bound* and the
    *upper_bound*.  These bins may or may not be equally divided
    in *log_space*, and the *lazy_reader* flag controls whether we
    use a memory conservative approach. If *end_collect* is True,
    take all values outside the given bounds and store them in the
    0 and *n_bins*-1 values.
    """
    def __init__(self, data_source, n_bins, bin_field,
                 lower_bound, upper_bound,
                 log_space = True, lazy_reader=False,
                 end_collect=False):
        BinnedProfile.__init__(self, data_source, lazy_reader)
        self.bin_field = bin_field
        self._x_log = log_space
        self.end_collect = end_collect
        self.n_bins = n_bins

        # Get our bins
        if log_space:
            func = np.logspace
            lower_bound, upper_bound = np.log10(lower_bound), np.log10(upper_bound)
        else:
            func = np.linspace

        # These are the bin *edges*
        self._bins = func(lower_bound, upper_bound, n_bins + 1)

        # These are the bin *left edges*.  These are the x-axis values
        # we plot in the PlotCollection
        self[bin_field] = self._bins

        # If we are not being memory-conservative, grab all the bins
        # and the inverse indices right now.
        if not lazy_reader:
            self._args = self._get_bins(data_source)

    def _get_empty_field(self):
        return np.zeros(self[self.bin_field].size, dtype='float64')

    @preserve_source_parameters
    def _bin_field(self, source, field, weight, accumulation,
                   args, check_cut=False):
        mi, inv_bin_indices = args # Args has the indices to use as input
        # check_cut is set if source != self._data_source
        # (i.e., lazy_reader)
        source_data = self._get_field(source, field, check_cut)
        if weight: weight_data = self._get_field(source, weight, check_cut)
        else: weight_data = np.ones(source_data.shape, dtype='float64')
        self.total_stuff = source_data.sum()
        binned_field = self._get_empty_field()
        weight_field = self._get_empty_field()
        m_field = self._get_empty_field()
        q_field = self._get_empty_field()
        used_field = self._get_empty_field()
        mi = args[0]
        bin_indices_x = args[1].ravel().astype('int64')
        source_data = source_data[mi]
        weight_data = weight_data[mi]
        bin_profile1d(bin_indices_x, weight_data, source_data,
                      weight_field, binned_field,
                      m_field, q_field, used_field)
        # Fix for laziness, because at the *end* we will be
        # summing up all of the histograms and dividing by the
        # weights.  Accumulation likely doesn't work with weighted
        # average fields.
        if accumulation: 
            binned_field = np.add.accumulate(binned_field)
        return binned_field, weight_field, q_field, \
            used_field.astype("bool")

    @preserve_source_parameters
    def _get_bins(self, source, check_cut=False):
        source_data = self._get_field(source, self.bin_field, check_cut)
        if source_data.size == 0: # Nothing for us here.
            raise EmptyProfileData()
        # Truncate at boundaries.
        if self.end_collect:
            mi = np.ones_like(source_data).astype('bool')
        else:
            mi = ((source_data > self._bins.min())
               &  (source_data < self._bins.max()))
        sd = source_data[mi]
        if sd.size == 0:
            raise EmptyProfileData()
        # Stick the bins into our fixed bins, set at initialization
        bin_indices = np.digitize(sd, self._bins)
        if self.end_collect: #limit the range of values to 0 and n_bins-1
            bin_indices = np.clip(bin_indices, 0, self.n_bins - 1)
        else: #throw away outside values
            bin_indices -= 1
          
        return (mi, bin_indices)

    def choose_bins(self, bin_style):
        # Depending on the bin_style, choose from bin edges 0...N either:
        # both: 0...N, left: 0...N-1, right: 1...N 
        # center: N bins that are the average (both in linear or log
        # space) of each pair of left/right edges
        x = self.field_data[self.bin_field]
        if bin_style is 'both': pass
        elif bin_style is 'left': x = x[:-1]
        elif bin_style is 'right': x = x[1:]
        elif bin_style is 'center':
            if self._x_log: x=np.log10(x)
            x = 0.5*(x[:-1] + x[1:])
            if self._x_log: x=10**x
        else:
            mylog.error('Did not recognize bin_style')
            raise ValueError
        return x

    def write_out(self, filename, format="%0.16e", bin_style='left'):
        ''' 
        Write out data in ascii file, using *format* and
        *bin_style* (left, right, center, both).
        '''
        fid = open(filename,"w")
        fields = [field for field in sorted(self.field_data.keys()) if field != "UsedBins"]
        fields.remove(self.bin_field)
        fid.write("\t".join(["#"] + [self.bin_field] + fields + ["\n"]))

        field_data = np.array(self.choose_bins(bin_style)) 
        if bin_style is 'both':
            field_data = np.append([field_data], np.array([self.field_data[field] for field in fields]), axis=0)
        else: 
            field_data = np.append([field_data], np.array([self.field_data[field][:-1] for field in fields]), axis=0)
        
        for line in range(field_data.shape[1]):
            field_data[:,line].tofile(fid, sep="\t", format=format)
            fid.write("\n")
        fid.close()

    def write_out_h5(self, filename, group_prefix=None, bin_style='left'):
        """
        Write out data in an hdf5 file *filename*.  Each profile is
        put into a group, named by the axis fields.  Optionally a
        *group_prefix* can be prepended to the group name.  If the
        group already exists, it will delete and replace.  However,
        due to hdf5 functionality, in only unlinks the data, so an
        h5repack may be necessary to conserve space.  Axes values are
        saved in group attributes.  Bins will be saved based on
        *bin_style* (left, right, center, both).
        """
        fid = h5py.File(filename)
        fields = [field for field in sorted(self.field_data.keys()) if (field != "UsedBins" and field != self.bin_field)]
        if group_prefix is None:
            name = "%s-1d" % (self.bin_field)
        else:
            name = "%s-%s-1d" % (group_prefix, self.bin_field)
            
        if name in fid: 
            mylog.info("Profile file is getting larger since you are attempting to overwrite a profile. You may want to repack")
            del fid[name] 
        group = fid.create_group(name)
        group.attrs["x-axis-%s" % self.bin_field] = self.choose_bins(bin_style)
        for field in fields:
            dset = group.create_dataset("%s" % field, data=self.field_data[field][:-1])
        fid.close()

    def _get_bin_fields(self):
        return [self.bin_field]

class BinnedProfile2D(BinnedProfile):
    """
    A 'Profile' produces either a weighted (or unweighted) average
    or a straight sum of a field in a bin defined by two other
    fields.  In the case of a weighted average, we have: p_i =
    sum( w_i * v_i ) / sum(w_i)

    We accept a *data_source*, which will be binned into
    *x_n_bins* by the field *x_bin_field* between the
    *x_lower_bound* and the *x_upper_bound* and then again binned
    into *y_n_bins* by the field *y_bin_field* between the
    *y_lower_bound* and the *y_upper_bound*.  These bins may or
    may not be equally divided in log-space as specified by
    *x_log* and *y_log*, and the *lazy_reader* flag controls
    whether we use a memory conservative approach. If
    *end_collect* is True, take all values outside the given
    bounds and store them in the 0 and *n_bins*-1 values.
    """
    def __init__(self, data_source,
                 x_n_bins, x_bin_field, x_lower_bound, x_upper_bound, x_log,
                 y_n_bins, y_bin_field, y_lower_bound, y_upper_bound, y_log,
                 lazy_reader=False, end_collect=False):
        BinnedProfile.__init__(self, data_source, lazy_reader)
        self.x_bin_field = x_bin_field
        self.y_bin_field = y_bin_field
        self._x_log = x_log
        self._y_log = y_log
        self.end_collect = end_collect
        self.x_n_bins = x_n_bins
        self.y_n_bins = y_n_bins

        func = {True:np.logspace, False:np.linspace}[x_log]
        bounds = fix_bounds(x_lower_bound, x_upper_bound, x_log)
        self._x_bins = func(bounds[0], bounds[1], x_n_bins + 1)
        self[x_bin_field] = self._x_bins

        func = {True:np.logspace, False:np.linspace}[y_log]
        bounds = fix_bounds(y_lower_bound, y_upper_bound, y_log)
        self._y_bins = func(bounds[0], bounds[1], y_n_bins + 1)
        self[y_bin_field] = self._y_bins

        if np.any(np.isnan(self[x_bin_field])) \
            or np.any(np.isnan(self[y_bin_field])):
            mylog.error("Your min/max values for x, y have given me a nan.")
            mylog.error("Usually this means you are asking for log, with a zero bound.")
            raise ValueError
        if not lazy_reader:
            self._args = self._get_bins(data_source)

    def _get_empty_field(self):
        return np.zeros((self[self.x_bin_field].size,
                         self[self.y_bin_field].size), dtype='float64')

    @preserve_source_parameters
    def _bin_field(self, source, field, weight, accumulation,
                   args, check_cut=False):
        source_data = self._get_field(source, field, check_cut)
        if weight: weight_data = self._get_field(source, weight, check_cut)
        else: weight_data = np.ones(source_data.shape, dtype='float64')
        self.total_stuff = source_data.sum()
        binned_field = self._get_empty_field()
        weight_field = self._get_empty_field()
        m_field = self._get_empty_field()
        q_field = self._get_empty_field()
        used_field = self._get_empty_field()
        mi = args[0]
        bin_indices_x = args[1].ravel().astype('int64')
        bin_indices_y = args[2].ravel().astype('int64')
        source_data = source_data[mi]
        weight_data = weight_data[mi]
        nx = bin_indices_x.size
        #mylog.debug("Binning %s / %s times", source_data.size, nx)
        bin_profile2d(bin_indices_x, bin_indices_y, weight_data, source_data,
                      weight_field, binned_field, m_field, q_field, used_field)
        if accumulation: # Fix for laziness
            if not iterable(accumulation):
                raise SyntaxError("Accumulation needs to have length 2")
            if accumulation[0]:
                binned_field = np.add.accumulate(binned_field, axis=0)
            if accumulation[1]:
                binned_field = np.add.accumulate(binned_field, axis=1)
        return binned_field, weight_field, q_field, \
            used_field.astype("bool")

    @preserve_source_parameters
    def _get_bins(self, source, check_cut=False):
        source_data_x = self._get_field(source, self.x_bin_field, check_cut)
        source_data_y = self._get_field(source, self.y_bin_field, check_cut)
        if source_data_x.size == 0:
            raise EmptyProfileData()

        if self.end_collect:
            mi = np.arange(source_data_x.size)
        else:
            mi = np.where( (source_data_x > self._x_bins.min())
                           & (source_data_x < self._x_bins.max())
                           & (source_data_y > self._y_bins.min())
                           & (source_data_y < self._y_bins.max()))
        sd_x = source_data_x[mi]
        sd_y = source_data_y[mi]
        if sd_x.size == 0 or sd_y.size == 0:
            raise EmptyProfileData()

        bin_indices_x = np.digitize(sd_x, self._x_bins) - 1
        bin_indices_y = np.digitize(sd_y, self._y_bins) - 1
        if self.end_collect:
            bin_indices_x = np.minimum(np.maximum(1, bin_indices_x), self.x_n_bins) - 1
            bin_indices_y = np.minimum(np.maximum(1, bin_indices_y), self.y_n_bins) - 1

        # Now we set up our inverse bin indices
        return (mi, bin_indices_x, bin_indices_y)

    def choose_bins(self, bin_style):
        # Depending on the bin_style, choose from bin edges 0...N either:
        # both: 0...N, left: 0...N-1, right: 1...N 
        # center: N bins that are the average (both in linear or log
        # space) of each pair of left/right edges

        x = self.field_data[self.x_bin_field]
        y = self.field_data[self.y_bin_field]
        if bin_style is 'both':
            pass
        elif bin_style is 'left':
            x = x[:-1]
            y = y[:-1]
        elif bin_style is 'right':
            x = x[1:]
            y = y[1:]
        elif bin_style is 'center':
            if self._x_log: x=np.log10(x)
            if self._y_log: y=np.log10(y)
            x = 0.5*(x[:-1] + x[1:])
            y = 0.5*(y[:-1] + y[1:])
            if self._x_log: x=10**x
            if self._y_log: y=10**y
        else:
            mylog.error('Did not recognize bin_style')
            raise ValueError

        return x,y

    def write_out(self, filename, format="%0.16e", bin_style='left'):
        """
        Write out the values of x,y,v in ascii to *filename* for every
        field in the profile.  Optionally a *format* can be specified.
        Bins will be saved based on *bin_style* (left, right, center,
        both).
        """
        fid = open(filename,"w")
        fields = [field for field in sorted(self.field_data.keys()) if field != "UsedBins"]
        fid.write("\t".join(["#"] + [self.x_bin_field, self.y_bin_field]
                          + fields + ["\n"]))
        x,y = self.choose_bins(bin_style)
        x,y = np.meshgrid(x,y)
        field_data = [x.ravel(), y.ravel()]
        if bin_style is not 'both':
            field_data += [self.field_data[field][:-1,:-1].ravel() for field in fields
                           if field not in [self.x_bin_field, self.y_bin_field]]
        else:
            field_data += [self.field_data[field].ravel() for field in fields
                           if field not in [self.x_bin_field, self.y_bin_field]]

        field_data = np.array(field_data)
        for line in range(field_data.shape[1]):
            field_data[:,line].tofile(fid, sep="\t", format=format)
            fid.write("\n")
        fid.close()

    def write_out_h5(self, filename, group_prefix=None, bin_style='left'):
        """
        Write out data in an hdf5 file.  Each profile is put into a
        group, named by the axis fields.  Optionally a group_prefix
        can be prepended to the group name.  If the group already
        exists, it will delete and replace.  However, due to hdf5
        functionality, in only unlinks the data, so an h5repack may be
        necessary to conserve space.  Axes values are saved in group
        attributes. Bins will be saved based on *bin_style* (left,
        right, center, both).
        """
        fid = h5py.File(filename)
        fields = [field for field in sorted(self.field_data.keys()) if (field != "UsedBins" and field != self.x_bin_field and field != self.y_bin_field)]
        if group_prefix is None:
            name = "%s-%s-2d" % (self.y_bin_field, self.x_bin_field)
        else:
            name = "%s-%s-%s-2d" % (group_prefix, self.y_bin_field, self.x_bin_field)
        if name in fid: 
            mylog.info("Profile file is getting larger since you are attempting to overwrite a profile. You may want to repack")
            del fid[name] 
        group = fid.create_group(name)

        xbins, ybins = self.choose_bins(bin_style)
        group.attrs["x-axis-%s" % self.x_bin_field] = xbins
        group.attrs["y-axis-%s" % self.y_bin_field] = ybins
        for field in fields:
            dset = group.create_dataset("%s" % field, data=self.field_data[field][:-1,:-1])
        fid.close()

    def _get_bin_fields(self):
        return [self.x_bin_field, self.y_bin_field]

def fix_bounds(upper, lower, logit):
    if logit: return np.log10(upper), np.log10(lower)
    return upper, lower

class BinnedProfile2DInlineCut(BinnedProfile2D):
    def __init__(self, data_source,
                 x_n_bins, x_bin_field, x_lower_bound, x_upper_bound, x_log,
                 y_n_bins, y_bin_field, y_lower_bound, y_upper_bound, y_log,
                 lazy_reader=False, end_collect=False):
        lazy_reader = False
        self.indices = data_source["Ones"].astype("bool")
        BinnedProfile2D.__init__(self, data_source,
                 x_n_bins, x_bin_field, x_lower_bound, x_upper_bound, x_log,
                 y_n_bins, y_bin_field, y_lower_bound, y_upper_bound, y_log,
                 lazy_reader, end_collect)

    @preserve_source_parameters
    def _bin_field(self, source, field, weight, accumulation,
                   args, check_cut=False):
        source_data = self._get_field(source, field, check_cut)
        if weight: weight_data = self._get_field(source, weight, check_cut)
        else: weight_data = np.ones(source_data.shape, dtype='float64')
        self.total_stuff = source_data.sum()
        binned_field = self._get_empty_field()
        weight_field = self._get_empty_field()
        used_field = self._get_empty_field()
        mi = args[0]
        bin_indices_x = args[1][self.indices].ravel().astype('int64')
        bin_indices_y = args[2][self.indices].ravel().astype('int64')
        source_data = source_data[mi][self.indices]
        weight_data = weight_data[mi][self.indices]
        nx = bin_indices_x.size
        #mylog.debug("Binning %s / %s times", source_data.size, nx)
        Bin2DProfile(bin_indices_x, bin_indices_y, weight_data, source_data,
                     weight_field, binned_field, used_field)
        if accumulation: # Fix for laziness
            if not iterable(accumulation):
                raise SyntaxError("Accumulation needs to have length 2")
            if accumulation[0]:
                binned_field = np.add.accumulate(binned_field, axis=0)
            if accumulation[1]:
                binned_field = np.add.accumulate(binned_field, axis=1)
        return binned_field, weight_field, used_field.astype('bool')

        
class BinnedProfile3D(BinnedProfile):
    """
    A 'Profile' produces either a weighted (or unweighted) average
    or a straight sum of a field in a bin defined by two other
    fields.  In the case of a weighted average, we have: p_i =
    sum( w_i * v_i ) / sum(w_i)
    
    We accept a *data_source*, which will be binned into
    *(x,y,z)_n_bins* by the field *(x,y,z)_bin_field* between the
    *(x,y,z)_lower_bound* and the *(x,y,z)_upper_bound*.  These bins may or
    may not be equally divided in log-space as specified by
    *(x,y,z)_log*, and the *lazy_reader* flag controls
    whether we use a memory conservative approach. If
    *end_collect* is True, take all values outside the given
    bounds and store them in the 0 and *n_bins*-1 values.
    """
    def __init__(self, data_source,
                 x_n_bins, x_bin_field, x_lower_bound, x_upper_bound, x_log,
                 y_n_bins, y_bin_field, y_lower_bound, y_upper_bound, y_log,
                 z_n_bins, z_bin_field, z_lower_bound, z_upper_bound, z_log,
                 lazy_reader=False, end_collect=False):
        BinnedProfile.__init__(self, data_source, lazy_reader)
        self.x_bin_field = x_bin_field
        self.y_bin_field = y_bin_field
        self.z_bin_field = z_bin_field
        self._x_log = x_log
        self._y_log = y_log
        self._z_log = z_log
        self.end_collect = end_collect
        self.x_n_bins = x_n_bins
        self.y_n_bins = y_n_bins
        self.z_n_bins = z_n_bins

        func = {True:np.logspace, False:np.linspace}[x_log]
        bounds = fix_bounds(x_lower_bound, x_upper_bound, x_log)
        self._x_bins = func(bounds[0], bounds[1], x_n_bins + 1)
        self[x_bin_field] = self._x_bins

        func = {True:np.logspace, False:np.linspace}[y_log]
        bounds = fix_bounds(y_lower_bound, y_upper_bound, y_log)
        self._y_bins = func(bounds[0], bounds[1], y_n_bins + 1)
        self[y_bin_field] = self._y_bins

        func = {True:np.logspace, False:np.linspace}[z_log]
        bounds = fix_bounds(z_lower_bound, z_upper_bound, z_log)
        self._z_bins = func(bounds[0], bounds[1], z_n_bins + 1)
        self[z_bin_field] = self._z_bins

        if np.any(np.isnan(self[x_bin_field])) \
            or np.any(np.isnan(self[y_bin_field])) \
            or np.any(np.isnan(self[z_bin_field])):
            mylog.error("Your min/max values for x, y or z have given me a nan.")
            mylog.error("Usually this means you are asking for log, with a zero bound.")
            raise ValueError
        if not lazy_reader:
            self._args = self._get_bins(data_source)

    def _get_empty_field(self):
        return np.zeros((self[self.x_bin_field].size,
                         self[self.y_bin_field].size,
                         self[self.z_bin_field].size), dtype='float64')

    @preserve_source_parameters
    def _bin_field(self, source, field, weight, accumulation,
                   args, check_cut=False):
        source_data = self._get_field(source, field, check_cut)
        weight_data = np.ones(source_data.shape).astype('float64')
        if weight: weight_data = self._get_field(source, weight, check_cut)
        else: weight_data = np.ones(source_data.shape).astype('float64')
        self.total_stuff = source_data.sum()
        binned_field = self._get_empty_field()
        weight_field = self._get_empty_field()
        m_field = self._get_empty_field()
        q_field = self._get_empty_field()
        used_field = self._get_empty_field()
        mi = args[0]
        bin_indices_x = args[1].ravel().astype('int64')
        bin_indices_y = args[2].ravel().astype('int64')
        bin_indices_z = args[3].ravel().astype('int64')
        source_data = source_data[mi]
        weight_data = weight_data[mi]
        bin_profile3d(bin_indices_x, bin_indices_y, bin_indices_z,
                      weight_data, source_data, weight_field, binned_field,
                      m_field, q_field, used_field)
        if accumulation: # Fix for laziness
            if not iterable(accumulation):
                raise SyntaxError("Accumulation needs to have length 2")
            if accumulation[0]:
                binned_field = np.add.accumulate(binned_field, axis=0)
            if accumulation[1]:
                binned_field = np.add.accumulate(binned_field, axis=1)
            if accumulation[2]:
                binned_field = np.add.accumulate(binned_field, axis=2)
        return binned_field, weight_field, q_field, \
            used_field.astype("bool")

    @preserve_source_parameters
    def _get_bins(self, source, check_cut=False):
        source_data_x = self._get_field(source, self.x_bin_field, check_cut)
        source_data_y = self._get_field(source, self.y_bin_field, check_cut)
        source_data_z = self._get_field(source, self.z_bin_field, check_cut)
        if source_data_x.size == 0:
            raise EmptyProfileData()
        if self.end_collect:
            mi = np.arange(source_data_x.size)
        else:
            mi = ( (source_data_x > self._x_bins.min())
                 & (source_data_x < self._x_bins.max())
                 & (source_data_y > self._y_bins.min())
                 & (source_data_y < self._y_bins.max())
                 & (source_data_z > self._z_bins.min())
                 & (source_data_z < self._z_bins.max()))
        sd_x = source_data_x[mi]
        sd_y = source_data_y[mi]
        sd_z = source_data_z[mi]
        if sd_x.size == 0 or sd_y.size == 0 or sd_z.size == 0:
            raise EmptyProfileData()

        bin_indices_x = np.digitize(sd_x, self._x_bins) - 1
        bin_indices_y = np.digitize(sd_y, self._y_bins) - 1
        bin_indices_z = np.digitize(sd_z, self._z_bins) - 1
        if self.end_collect:
            bin_indices_x = np.minimum(np.maximum(1, bin_indices_x), self.x_n_bins) - 1
            bin_indices_y = np.minimum(np.maximum(1, bin_indices_y), self.y_n_bins) - 1
            bin_indices_z = np.minimum(np.maximum(1, bin_indices_z), self.z_n_bins) - 1

        # Now we set up our inverse bin indices
        return (mi, bin_indices_x, bin_indices_y, bin_indices_z)

    def choose_bins(self, bin_style):
        # Depending on the bin_style, choose from bin edges 0...N either:
        # both: 0...N, left: 0...N-1, right: 1...N 
        # center: N bins that are the average (both in linear or log
        # space) of each pair of left/right edges

        x = self.field_data[self.x_bin_field]
        y = self.field_data[self.y_bin_field]
        z = self.field_data[self.z_bin_field]
        if bin_style is 'both':
            pass
        elif bin_style is 'left':
            x = x[:-1]
            y = y[:-1]
            z = z[:-1]
        elif bin_style is 'right':
            x = x[1:]
            y = y[1:]
            z = z[1:]
        elif bin_style is 'center':
            if self._x_log: x=np.log10(x)
            if self._y_log: y=np.log10(y)
            if self._z_log: z=np.log10(z)
            x = 0.5*(x[:-1] + x[1:])
            y = 0.5*(y[:-1] + y[1:])
            z = 0.5*(z[:-1] + z[1:])
            if self._x_log: x=10**x
            if self._y_log: y=10**y
            if self._z_log: y=10**z
        else:
            mylog.error('Did not recognize bin_style')
            raise ValueError

        return x,y,z

    def write_out(self, filename, format="%0.16e"):
        pass # Will eventually dump HDF5

    def write_out_h5(self, filename, group_prefix=None, bin_style='left'):
        """
        Write out data in an hdf5 file.  Each profile is put into a
        group, named by the axis fields.  Optionally a group_prefix
        can be prepended to the group name.  If the group already
        exists, it will delete and replace.  However, due to hdf5
        functionality, in only unlinks the data, so an h5repack may be
        necessary to conserve space.  Axes values are saved in group
        attributes.
        """
        fid = h5py.File(filename)
        fields = [field for field in sorted(self.field_data.keys()) 
                  if (field != "UsedBins" and field != self.x_bin_field and field != self.y_bin_field and field != self.z_bin_field)]
        if group_prefix is None:
            name = "%s-%s-%s-3d" % (self.z_bin_field, self.y_bin_field, self.x_bin_field)
        else:
            name = "%s-%s-%s-%s-3d" % (group_prefix,self.z_bin_field, self.y_bin_field, self.x_bin_field)

        if name in fid: 
            mylog.info("Profile file is getting larger since you are attempting to overwrite a profile. You may want to repack")
            del fid[name]
        group = fid.create_group(name)

        xbins, ybins, zbins= self.choose_bins(bin_style)
        group.attrs["x-axis-%s" % self.x_bin_field] = xbins
        group.attrs["y-axis-%s" % self.y_bin_field] = ybins
        group.attrs["z-axis-%s" % self.z_bin_field] = zbins
        
        for field in fields:
            dset = group.create_dataset("%s" % field, data=self.field_data[field][:-1,:-1,:-1])
        fid.close()


    def _get_bin_fields(self):
        return [self.x_bin_field, self.y_bin_field, self.z_bin_field]

    def store_profile(self, name, force=False):
        """
        By identifying the profile with a fixed, user-input *name* we can
        store it in the serialized data section of the hierarchy file.  *force*
        governs whether or not an existing profile with that name will be
        overwritten.
        """
        # First we get our data in order
        order = []
        set_attr = {'x_bin_field':self.x_bin_field,
                    'y_bin_field':self.y_bin_field,
                    'z_bin_field':self.z_bin_field,
                    'x_bin_values':self[self.x_bin_field],
                    'y_bin_values':self[self.y_bin_field],
                    'z_bin_values':self[self.z_bin_field],
                    '_x_log':self._x_log,
                    '_y_log':self._y_log,
                    '_z_log':self._z_log,
                    'shape': (self[self.x_bin_field].size,
                              self[self.y_bin_field].size,
                              self[self.z_bin_field].size),
                    'field_order':order }
        values = []
        for field in self.field_data:
            if field in set_attr.values(): continue
            order.append(field)
            values.append(self[field].ravel())
        values = np.array(values).transpose()
        self._data_source.hierarchy.save_data(values, "/Profiles", name,
                                              set_attr, force=force)

class StoredBinnedProfile3D(BinnedProfile3D):
    def __init__(self, pf, name):
        """
        Given a *pf* parameterfile and the *name* of a stored profile, retrieve
        it into a read-only data structure.
        """
        self.field_data = YTFieldData()
        prof_arr = pf.h.get_data("/Profiles", name)
        if prof_arr is None: raise KeyError("No such array")
        for ax in 'xyz':
            for base in ['%s_bin_field', '_%s_log']:
                setattr(self, base % ax, prof_arr.getAttr(base % ax))
        for ax in 'xyz':
            fn = getattr(self, '%s_bin_field' % ax)
            self.field_data[fn] = prof_arr.getAttr('%s_bin_values' % ax)
        shape = prof_arr.getAttr('shape')
        for fn, fd in zip(prof_arr.getAttr('field_order'),
                          prof_arr.read().transpose()):
            self.field_data[fn] = fd.reshape(shape)

    def add_fields(self, *args, **kwargs):
        raise RuntimeError("Sorry, you can't add to a stored profile.")



"""
Logging facility for yt
Will initialize everything, and associate one with each module

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2013 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from yt.config import ytcfg

# This next bit is grabbed from:
# http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if(levelno >= 50):
            color = '\x1b[31m'  # red
        elif(levelno >= 40):
            color = '\x1b[31m'  # red
        elif(levelno >= 30):
            color = '\x1b[33m'  # yellow
        elif(levelno >= 20):
            color = '\x1b[32m'  # green
        elif(levelno >= 10):
            color = '\x1b[35m'  # pink
        else:
            color = '\x1b[0m'  # normal
        ln = color + args[1].levelname + '\x1b[0m'
        args[1].levelname = ln
        return fn(*args)
    return new

level = min(max(ytcfg.getint("yt", "loglevel"), 0), 50)
ufstring = "%(name)-3s: [%(levelname)-9s] %(asctime)s %(message)s"
cfstring = "%(name)-3s: [%(levelname)-18s] %(asctime)s %(message)s"
logging.basicConfig(
    format=ufstring,
    level=level
)

rootLogger = logging.getLogger()
ytLogger = logging.getLogger("yt")


def disable_stream_logging():
    # We just remove the root logger's handlers
    for handler in rootLogger.handlers:
        if isinstance(handler, logging.StreamHandler):
            rootLogger.removeHandler(handler)
    h = logging.NullHandler()
    ytLogger.addHandler(h)

original_emitter = logging.StreamHandler.emit


def colorize_logging():
    f = logging.Formatter(cfstring)
    if len(rootLogger.handlers) > 0:
        rootLogger.handlers[0].setFormatter(f)
    logging.StreamHandler.emit = add_coloring_to_emit_ansi(
        logging.StreamHandler.emit)


def uncolorize_logging():
    f = logging.Formatter(ufstring)
    if len(rootLogger.handlers) > 0:
        rootLogger.handlers[0].setFormatter(f)
    logging.StreamHandler.emit = original_emitter

if ytcfg.getboolean("yt", "coloredlogs"):
    colorize_logging()

if ytcfg.getboolean("yt", "suppressStreamLogging"):
    disable_stream_logging()

ytLogger.debug("Set log level to %s", level)



"""
API for yt.data_objects

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: J.S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: MSU
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from grid_patch import \
    AMRGridPatch

from hierarchy import \
    AMRHierarchy

from static_output import \
    StaticOutput

from object_finding_mixin import \
    ObjectFindingMixin

from particle_io import \
    ParticleIOHandler, \
    particle_handler_registry

from profiles import \
    EmptyProfileData, \
    BinnedProfile, \
    BinnedProfile1D, \
    BinnedProfile2D, \
    BinnedProfile3D

from time_series import \
    TimeSeriesData, \
    TimeSeriesDataObject

from analyzer_objects import \
    AnalysisTask, analysis_task

from data_containers import \
    data_object_registry

from derived_quantities import \
    quantity_info, \
    add_quantity

from image_array import \
    ImageArray

from field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    NeedsGridType, \
    NeedsOriginalGrid, \
    NeedsDataField, \
    NeedsProperty, \
    NeedsParameter, \
    FieldDetector, \
    DerivedField, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType, \
    add_field, \
    add_grad, \
    derived_field

from particle_trajectories import \
    ParticleTrajectoryCollection



#!/usr/bin/env python

from os.path import join

def configuration(parent_package = '', top_path = None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
    config = Configuration('spatial', parent_package, top_path)

    config.add_data_dir('tests')
    config.add_sconscript('SConstruct')

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(maintainer = "SciPy Developers",
          author = "Anne Archibald",
          maintainer_email = "scipy-dev@scipy.org",
          description = "Spatial algorithms and data structures",
          url = "http://www.scipy.org",
          license = "SciPy License (BSD Style)",
          **configuration(top_path='').todict()
          )



"""
A read-eval-print-loop.  This code was released in the CherryPy project
initially, but has been heavily modified and again re-released in compliance
with the terms of its original license.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import codeop
import inspect
import re
import json
import sys
import traceback
from cStringIO import StringIO

class ProgrammaticREPL(object):
    stopped = False
    debug = False
    def __init__(self, locals=None):
        self.executed_cell_texts = []
        self.locals = {}
        if locals:
            self.locals.update(locals)
        self.buffer = []
        # Nominally at this point we could populate our namespace with widgets
        # or other useful functions.  We aren't really ready for that just yet.

    def evaluate_cell(self, input):
        result = []
        for line in input.split("\n"):
            r = self.push(line)
            if r is not None: result.append(r)
        # To catch if people leave a little bit at the end for us
        if r is None:
            r = self.push("\n")
            if r is not None: result.append(r)
        # Usually they have \n already
        return "".join(result)
    
    def push(self, line):
        """Push 'line' and return exec results (None if more input needed)."""
        if line == "help":
            return "Type help(object) for help about object."
        if line == "help()":
            return "You cannot call help() without an argument."
        
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        
        try:
            code = codeop.compile_command(source, "<HTTP input>", 'single')
        except (OverflowError, SyntaxError, ValueError):
            self.buffer = []
            return traceback.format_exc()
        
        if code is None:
            # More lines needed.
            return None
        
        self.buffer = []
        return self.execute(code)
    
    def execute(self, code):
        """Execute the given code in self.locals and return any stdout/sterr."""
        out = StringIO()
        oldout = sys.stdout
        olderr = sys.stderr
        sys.stdout = sys.stderr = out
        try:
            try:
                exec code in self.locals
            except:
                result = traceback.format_exc()
            else:
                result = out.getvalue()
        finally:
            sys.stdout = oldout
            sys.stderr = olderr
        out.close()
        return result
    
    def dir(self, line):
        """Examine a partial line and provide attr list of final expr."""
        line = re.split(r"\s", line)[-1].strip()
        # Support lines like "thing.attr" as "thing.", because the browser
        # may not finish calculating the partial line until after the user
        # has clicked on a few more keys.
        line = ".".join(line.split(".")[:-1])
        try:
            result = eval("dir(%s)" % line, {}, self.locals)
        except:
            return []
        return result
    
    def doc(self, line):
        """Examine a partial line and provide sig+doc of final expr."""
        line = re.split(r"\s", line)[-1].strip()
        # Support lines like "func(text" as "func(", because the browser
        # may not finish calculating the partial line until after the user
        # has clicked on a few more keys.
        line = "(".join(line.split("(")[:-1])
        try:
            result = eval(line, {}, self.locals)
            try:
                if isinstance(result, type):
                    func = result.__init__
                else:
                    func = result
                args, varargs, varkw, defaults = inspect.getargspec(func)
            except TypeError:
                if callable(result):
                    doc = getattr(result, "__doc__", "") or ""
                    return "%s\n\n%s" % (line, doc)
                return None
        except:
            return None
        
        if args and args[0] == 'self':
            args.pop(0)
        missing = object()
        defaults = defaults or []
        defaults = ([missing] * (len(args) - len(defaults))) + list(defaults)
        arglist = []
        for a, d in zip(args, defaults):
            if d is missing:
                arglist.append(a)
            else:
                arglist.append("%s=%s" % (a, d))
        if varargs:
            arglist.append("*%s" % varargs)
        if varkw:
            arglist.append("**%s" % varkw)
        doc = getattr(result, "__doc__", "") or ""
        return "%s(%s)\n%s" % (line, ", ".join(arglist), doc)



"""
Minimalist performance counting for yt

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.config import ytcfg
from yt.funcs import *
import time
from datetime import datetime as dt
from bisect import insort
import atexit

class PerformanceCounters(object):
    _shared_state = {}
    def __new__(cls, *args, **kwargs):
        self = object.__new__(cls, *args, **kwargs)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self):
        self.counters = defaultdict(lambda: 0.0)
        self.counting = defaultdict(lambda: False)
        self.starttime = defaultdict(lambda: 0)
        self.endtime = defaultdict(lambda: 0)
        self._on = ytcfg.getboolean("yt", "timefunctions")
        self.exit()

    def __call__(self, name):
        if not self._on: return
        if self.counting[name]:
            self.counters[name] = time.time() - self.counters[name] 
            self.counting[name] = False
            self.endtime[name] = dt.now()
        else:
            self.counters[name] = time.time()
            self.counting[name] = True
            self.starttime[name] = dt.now()

    def call_func(self, func):
        if not self._on: return func
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            self(func.func_name)
            func(*args, **kwargs)
            self(func.func_name)
        return func_wrapper

    def print_stats(self):
        mylog.info("Current counter status:\n")
        times = []
        for i in self.counters:
            insort(times, [self.starttime[i], i, 1]) # 1 for 'on'
            if not self.counting[i]:
                insort(times, [self.endtime[i], i, 0]) # 0 for 'off'
        shifts = {}
        order = []
        endtimes = {}
        shift = 0
        multi = 5
        for i in times:
            # a starting entry
            if i[2] == 1:
                shifts[i[1]] = shift
                order.append(i[1])
                shift += 1
            if i[2] == 0:
                shift -= 1
                endtimes[i[1]] = self.counters[i[1]]
        line = ''
        for i in order:
            if self.counting[i]:
                line = "%s%s%i : %s : still running\n" % (line, " "*shifts[i]*multi, shifts[i], i)
            else:
                line = "%s%s%i : %s : %0.3e\n" % (line, " "*shifts[i]*multi, shifts[i], i, self.counters[i])
        mylog.info("\n" + line)

    def exit(self):
        if self._on:
            atexit.register(self.print_stats)

yt_counters = PerformanceCounters()
time_function = yt_counters.call_func


class ProfilingController(object):
    def __init__(self):
        self.profilers = {}

    def profile_function(self, function_name):
        def wrapper(func):
            try:
                import cProfile
            except ImportError:
                return func
            my_prof = cProfile.Profile()
            self.profilers[function_name] = my_prof
            @wraps(func)
            def run_in_profiler(*args, **kwargs):
                my_prof.enable()
                func(*args, **kwargs)
                my_prof.disable()
            return run_in_profiler
        return wrapper

    def write_out(self, filename_prefix):
        if ytcfg.getboolean("yt","__parallel"):
            pfn = "%s_%03i_%03i" % (filename_prefix,
                     ytcfg.getint("yt", "__global_parallel_rank"),
                    ytcfg.getint("yt", "__global_parallel_size"))
        else:
            pfn = "%s" % (filename_prefix)
        for n, p in sorted(self.profilers.items()):
            fn = "%s_%s.cprof" % (pfn, n)
            mylog.info("Dumping %s into %s", n, fn)
            p.dump_stats(fn)



"""
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation:  UCSD
License:
  Copyright (C) 2010-2011 Matthew Turk  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# A simple image viewer, soon to be useful for stereo images, using OpenGL.

import sys
import OpenGL.GL as GL
import OpenGL.GLUT as GLUT
import OpenGL.GLU as GLU
import OpenGL.GL.shaders as shaders
from OpenGL.arrays import vbo, ArrayDatatype
import Image
import glob
import numpy as np
import time

ESCAPE = '\033'

class ViewHandler3D(object):
    def __init__(self, scene):
        # We 
        self.scene = scene
        self.dispatch_table = dict(
            q = (scene.translate, (1,  1.0)),
            e = (scene.translate, (1, -1.0)),
            w = (scene.translate, (2,  1.0)),
            s = (scene.translate, (2, -1.0)),
            a = (scene.translate, (0,  1.0)),
            d = (scene.translate, (0, -1.0)),

            Q = (scene.rotate, (1,  1.0)),
            E = (scene.rotate, (1, -1.0)),
            W = (scene.rotate, (2,  1.0)),
            S = (scene.rotate, (2, -1.0)),
            A = (scene.rotate, (0,  1.0)),
            D = (scene.rotate, (0, -1.0)),

            ESCAPE = (sys.exit, (0,))
        )

    def __call__(self, *args):
        # We set up our standard handlers, and then anything additional can get
        # called if none of our dispatch mechanisms work.
        if args[0] in self.dispatch_table:
            func, args = self.dispatch_table[args[0]]
            func(*args)
        # always draw when handling a keypress, even if it's one time too many
        self.scene.draw() 

class GenericGLUTScene(object):
    
    def __init__(self, width, height):
        self.init_glut(width, height)
        self.init_opengl(width, height)

    def init_glut(self, width, height):
        GLUT.glutInit([]) # drop sys.argv
        GLUT.glutInitDisplayMode(self._display_mode)
        GLUT.glutInitWindowSize(width, height)
        GLUT.glutInitWindowPosition(0, 0)
        self.window = GLUT.glutCreateWindow(self._title)
        GLUT.glutDisplayFunc(self.draw)
        #GLUT.glutIdleFunc(self.draw)
        GLUT.glutKeyboardFunc(self.keypress_handler)

    def run(self):
        GLUT.glutMainLoop()

class MultiImageDisplayScene(object):
    _display_mode = (GLUT.GLUT_RGBA | GLUT.GLUT_DOUBLE | GLUT.GLUT_DEPTH)
    _title = "Image Display"
    def __init__(self):
        GenericGLUTScene(512, 512)
        self._frames = []
        self._current = -1

    def add_image(self, obj):
        self._frames.append(obj)

    def init_opengl(self, width, height):
        GL.glClearColor(0.0, 0.0, 0.0, 0.0)
        GL.glClearDepth(1.0)
        GL.glDepthFunc(GL.GL_LESS)
        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glShadeModel(GL.GL_SMOOTH)
        
    def draw(self):

        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        if self._current >= 0:
            self._draw_current()

        GLUT.glutSwapBuffers()

    def _draw_current(self):
        self._frames[self._current].draw()

    def keypress_handler(self, *args):
        if args[0] == ESCAPE:
            sys.exit()
        elif len(self._frames) == 0:
            # The rest only operate if we have multiple frames
            return
        elif args[0] == 'n':
            self._current = min(self._current + 1, len(self._frames))
        elif args[0] == 'N':
            while self._current < len(self._frames) - 1:
                self._current += 1
                self.draw()
        elif args[0] == 'p':
            self._current = max(self._current - 1, 0)
        elif args[0] == 'P':
            while self._current > 0:
                self._current -= 1
                self.draw()
        elif args[0] == 't':
            for i in xrange(15):
                time.sleep(0.05)
                self.draw()
                self._current += 1
                time.sleep(0.05)
                self.draw()
                self._current -= 1
        self.draw() # Once more for good measure

class StereoMultiImageDisplayScene(MultiImageDisplayScene):
    _display_mode = (GLUT.GLUT_RGBA | GLUT.GLUT_DOUBLE | GLUT.GLUT_DEPTH |
                     GLUT.GLUT_STEREO)

    def draw(self):
        GL.glDrawBuffer(GL.GL_BACK_LEFT)

        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        if self._current >= 0:
            self._draw_current_left()

        GLUT.glutSwapBuffers()

        GL.glDrawBuffer(GL.GL_BACK_RIGHT)

        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        if self._current >= 0:
            self._draw_current_right()

        GLUT.glutSwapBuffers()

    def _draw_current_left(self):
        self._frames[self._current].draw_left()

    def _draw_current_right(self):
        self._frames[self._current].draw_right()

class FlatImage(object):
    def __init__(self, tex_unit = GL.GL_TEXTURE0):
        self.tex_unit = tex_unit
        GL.glActiveTexture(self.tex_unit)
        self._id = GL.glGenTextures(1)

    def draw(self):
        GL.glActiveTexture(self.tex_unit)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._id)
        GL.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_DECAL)

        GL.glColor3f(0.3, 0.5, 1.0)

        GL.glBegin(GL.GL_QUADS)
        GL.glVertex3i(-1, -1, -1)
        GL.glTexCoord2i(0, 0)

        GL.glVertex3i( 1, -1, -1)
        GL.glTexCoord2i(0, 1)

        GL.glVertex3i( 1,  1, -1)
        GL.glTexCoord2i(1, 1)

        GL.glVertex3i(-1,  1, -1)
        GL.glTexCoord2i(1, 0)
        GL.glEnd()

    def upload_image(self, buffer):
        ix, iy, nc = buffer.shape
        if nc != 4: raise RuntimeError
        GL.glActiveTexture(self.tex_unit)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._id)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, ix, iy, 0, GL.GL_RGBA,
                        GL.GL_UNSIGNED_BYTE, buffer.tostring())
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)

    @classmethod
    def from_image_file(cls, fn, tex_unit = GL.GL_TEXTURE0):
        buffer = np.array(Image.open(fn))
        print "Uploading buffer", buffer.min(), buffer.max(), buffer.shape, buffer.dtype
        obj = cls(tex_unit)
        obj.upload_image(buffer)
        return obj

class StereoImagePair(FlatImage):
    def __init__(self, tex_unit = GL.GL_TEXTURE0):
        self.tex_unit = tex_unit
        self.left_image = FlatImage(tex_unit)
        self.right_image = FlatImage(tex_unit)

    def draw_left(self):
        self.left_image.draw()

    def draw_right(self):
        self.right_image.draw()

    def upload_images(self, buffer_left, buffer_right):
        self.left_image.upload_image(buffer_left)
        self.right_image.upload_image(buffer_right)

    @classmethod
    def from_image_files(cls, left_fn, right_fn, tex_unit = GL.GL_TEXTURE0):
        print "Uploading pairs from %s and %s" % (left_fn, right_fn)
        left_buffer = np.array(Image.open(left_fn))
        right_buffer = np.array(Image.open(right_fn))
        obj = cls(tex_unit)
        obj.left_image.upload_image(left_buffer)
        obj.right_image.upload_image(right_buffer)
        return obj

_verts = ( (0,0,0), (1,0,0), (1,1,0), (0,1,0),
           (0,0,0), (1,0,0), (1,0,1), (0,0,1),
           (0,0,0), (0,0,1), (0,1,1), (0,1,0),
           (0,1,0), (1,1,0), (1,1,1), (0,1,1),
           (1,1,0), (1,0,0), (1,0,1), (1,1,1),
           (0,0,1), (0,1,1), (1,1,1), (1,0,1) )

class GridObject3DScene(GenericGLUTScene):
    _display_mode = (GLUT.GLUT_RGBA | GLUT.GLUT_DOUBLE | GLUT.GLUT_DEPTH)
    _title = "Grids"

    def _get_grid_vertices(self, offset):
        DLE, DRE = self.pf.domain_left_edge, pf.domain_right_edge
        DW = DRE - DLE
        k = 0
        for g in self.pf.h.grids:
            vs = ((g.LeftEdge-DLE)/DW, (g.RightEdge-DLE)/DW)
            for vert in _verts:
                for i,v in enumerate(vert):
                    yield vs[v][i] - offset
                    k += 1

    def __init__(self, pf, offset = 0.5):
        self.pf = pf
        GenericGLUTScene.__init__(self, 800, 800)

        num = len(pf.h.grids) * 6 * 4
        self.v = np.fromiter(self._get_grid_vertices(offset),
                             dtype = 'float32', count = num * 3)

        self.vertices = vbo.VBO(self.v)
        self.ng = len(pf.h.grids)
        self.ox = self.oy = self.rx = self.ry = self.rz = 0
        self.oz = -4

    def init_opengl(self, width, height):
        # One-time GL setup
        GL.glClearColor(1, 1, 1, 1)
        GL.glColor3f(1, 0, 0)
        GL.glEnable(GL.GL_DEPTH_TEST)
        #glEnable(GL_CULL_FACE)

        # Uncomment this line for a wireframe view
        GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE)

        # Simple light setup.  On Windows GL_LIGHT0 is enabled by default,
        # but this is not the case on Linux or Mac, so remember to always 
        # include it.
        GL.glEnable(GL.GL_LIGHTING)

        def vec(*args):
            return (GL.GLfloat * len(args))(*args)

        GL.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, vec(.5, .5, 1, 0))
        GL.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, vec(.5, .5, 1, 1))
        GL.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, vec(1, 1, 1, 1))
        GL.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, vec(1, 0, .5, 0))
        GL.glLightfv(GL.GL_LIGHT1, GL.GL_DIFFUSE, vec(.5, .5, .5, 1))
        GL.glLightfv(GL.GL_LIGHT1, GL.GL_SPECULAR, vec(1, 1, 1, 1))

        GL.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE, vec(0.5, 0, 0.3, 1))
        GL.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, vec(1, 1, 1, 1))
        GL.glMaterialf(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, 50)

        GL.glViewport(0, 0, width, height)
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GLU.gluPerspective(60., width / float(height), 1e-3, 10.)
        GL.glMatrixMode(GL.GL_MODELVIEW)
        
    def draw(self):

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        GL.glLoadIdentity()
        GL.glTranslatef(self.ox, self.oy, self.oz)
        GL.glRotatef(self.rx, 0, 0, 1)
        GL.glRotatef(self.ry, 0, 1, 0)
        GL.glRotatef(self.rz, 1, 0, 0)

        self.vertices.bind()
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glVertexPointer( 3, GL.GL_FLOAT, 0, self.vertices)
        GL.glDrawArrays(GL.GL_QUADS, 0, 4*6*self.ng)
        
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        self.vertices.unbind()
        GLUT.glutSwapBuffers()
        
    def keypress_handler(self, *args):
        tfac = 25.0
        rfac = 0.5
        if args[0] == ESCAPE:
            sys.exit()
        elif args[0] == 'a':
            self.ox += 1.0/tfac
        elif args[0] == 'd':
            self.ox -= 1.0/tfac
        elif args[0] == 's':
            self.oz -= 1.0/tfac
        elif args[0] == 'w':
            self.oz += 1.0/tfac
        elif args[0] == 'q':
            self.oy -= 1.0/tfac
        elif args[0] == 'e':
            self.oy += 1.0/tfac
        # Now, rotations
        elif args[0] == 'A':
            self.rx -= 1.0/rfac
        elif args[0] == 'D':
            self.rx += 1.0/rfac
        elif args[0] == 'S':
            self.rz -= 1.0/rfac
        elif args[0] == 'W':
            self.rz += 1.0/rfac
        elif args[0] == 'Q':
            self.ry -= 1.0/rfac
        elif args[0] == 'E':
            self.ry += 1.0/rfac
        self.draw()

class GridSlice3DScene(GenericGLUTScene):
    _display_mode = (GLUT.GLUT_RGBA | GLUT.GLUT_DOUBLE | GLUT.GLUT_DEPTH)
    _title = "Grids"

    def _get_grid_vertices(self, offset):
        for g in self.pf.h.grids:
            vs = (g.LeftEdge, g.RightEdge)
            for vert in _verts:
                for i,v in enumerate(vert):
                    yield vs[v][i] - offset

    def _setup_grids(self):
        self._grid_textures = {}
        for g in self.pf.h.grids:
            self._upload_grid_textures(g)

    def _upload_grid_textures(self, grid):
        ix, iy, iz = grid.ActiveDimensions

        GL.glActiveTexture(GL.GL_TEXTURE0)
        id_field = GL.glGenTextures(1)
        upload = np.log10(grid["Density"].astype("float32")).copy()
        self.mi = min(upload.min(), self.mi)
        self.ma = max(upload.max(), self.ma)
        #upload = (255*(upload - -31.0) / (-25.0 - -31.0)).astype("uint8")
        
        GL.glBindTexture(GL.GL_TEXTURE_3D, id_field)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexImage3D(GL.GL_TEXTURE_3D, 0, GL.GL_LUMINANCE32F_ARB, iz, iy, ix, 0,
                        GL.GL_LUMINANCE, GL.GL_FLOAT, upload)

        GL.glActiveTexture(GL.GL_TEXTURE1)
        id_mask  = GL.glGenTextures(1)
        upload = grid.child_mask.astype("float32").copy()

        GL.glBindTexture(GL.GL_TEXTURE_3D, id_mask)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, GL.GL_CLAMP)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexImage3D(GL.GL_TEXTURE_3D, 0, GL.GL_LUMINANCE, iz, iy, ix, 0,
                        GL.GL_LUMINANCE, GL.GL_FLOAT, upload)

        self._grid_textures[grid.id] = (id_field, id_mask)

        print "Uploaded", grid.id

    def __init__(self, pf, offset = 0.5):
        self.offset = offset
        self.mi, self.ma = 1e30, -1e30
        self.pf = pf
        self.coord = 0.0
        self.tfac = 10.0
        self.rfac = 0.5
        self._setup_keypress_handler()
        GenericGLUTScene.__init__(self, 800, 800)

        num = len(pf.h.grids) * 6 * 4
        self.v = np.fromiter(self._get_grid_vertices(offset),
                             dtype = 'float32', count = num * 3)

        self.vertices = vbo.VBO(self.v)
        self.ng = len(pf.h.grids)
        self.position = np.zeros(3, dtype='float')
        self.rotation = np.zeros(3, dtype='float')
        self.position[2] = -2 # Offset backwards a bit

        self._setup_grids()

    def init_opengl(self, width, height):
        # One-time GL setup
        GL.glClearColor(1, 1, 1, 1)
        GL.glColor3f(1, 0, 0)
        GL.glEnable(GL.GL_DEPTH_TEST)

        GL.glViewport(0, 0, width, height)
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GLU.gluPerspective(60., width / float(height), 1e-3, 10.)
        GL.glMatrixMode(GL.GL_MODELVIEW)

        # Now we compile our shaders

        self.program = shaders.compileProgram(
            shaders.compileShader('''
                void main() {
                    gl_TexCoord[0]=gl_TextureMatrix[0] * gl_MultiTexCoord0;
                    gl_TexCoord[1]=gl_TextureMatrix[1] * gl_MultiTexCoord1;
                    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
                }
            ''',GL.GL_VERTEX_SHADER),
            shaders.compileShader('''
                uniform float ma;
                uniform float mi;
                uniform sampler3D field;
                uniform sampler3D mask;

                void main() {
                    vec3 pos;
                    float val;

                    pos = vec3(gl_TexCoord[1].xyz);
                    val = texture3D( mask, pos )[0];
                    if(val == 0.0) discard;

                    pos = vec3(gl_TexCoord[0].xyz);
                    val = texture3D( field, pos )[0];

                    float color = (val - mi) / (ma - mi);
                
                    //gl_FragColor = vec4(pos.x, pos.y, pos.z, 1.0);
                    gl_FragColor = vec4(color, color, color, 1.0);
                }
        ''',GL.GL_FRAGMENT_SHADER),)

        self.uniform_locations = dict( (
                (v, GL.glGetUniformLocation(self.program, v)) for v in
                ['mi','ma','field','mask']
            ) )
            
    def draw(self):

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE)

        GL.glLoadIdentity()
        GL.glTranslatef(*self.position)
        GL.glRotatef(self.rotation[0], 0, 0, 1)
        GL.glRotatef(self.rotation[1], 0, 1, 0)
        GL.glRotatef(self.rotation[2], 1, 0, 0)

        self.vertices.bind()
        GL.glColor3f(0.0, 0.0, 0.0)
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glVertexPointer( 3, GL.GL_FLOAT, 0, self.vertices)
        GL.glDrawArrays(GL.GL_QUADS, 0, 4*6*self.ng)
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        self.vertices.unbind()

        # Now, we just want to draw 
        GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL)

        # We can just draw a single quad for now
        GL.glUseProgram(self.program)
        GL.glUniform1i(self.uniform_locations["field"], 0)
        GL.glUniform1i(self.uniform_locations["mask"], 1)
        GL.glUniform1f(self.uniform_locations["mi"], self.mi)
        GL.glUniform1f(self.uniform_locations["ma"], self.ma)
        GL.glEnable(GL.GL_TEXTURE_3D)

        t0, t1 = 0.0, 1.0
        for g in self.pf.h.find_slice_grids(self.coord + 0.5, 1)[0]:
            LE = g.LeftEdge - self.offset
            RE = g.RightEdge - self.offset
            off = (self.coord - LE[1]) / (RE[1] - LE[1])

            GL.glActiveTexture(GL.GL_TEXTURE0)
            GL.glBindTexture(GL.GL_TEXTURE_3D, self._grid_textures[g.id][0])
            GL.glEnable(GL.GL_TEXTURE_3D)

            GL.glActiveTexture(GL.GL_TEXTURE1)
            GL.glBindTexture(GL.GL_TEXTURE_3D, self._grid_textures[g.id][1])
            GL.glEnable(GL.GL_TEXTURE_3D)

            GL.glBegin(GL.GL_QUADS)

            GL.glMultiTexCoord3f(GL.GL_TEXTURE0, t0, off, t0)
            GL.glMultiTexCoord3f(GL.GL_TEXTURE1, t0, off, t0)
            GL.glVertex3f(LE[0], self.coord, LE[2])

            GL.glMultiTexCoord3f(GL.GL_TEXTURE0, t0, off, t1)
            GL.glMultiTexCoord3f(GL.GL_TEXTURE1, t0, off, t1)
            GL.glVertex3f(RE[0], self.coord, LE[2])

            GL.glMultiTexCoord3f(GL.GL_TEXTURE0, t1, off, t1)
            GL.glMultiTexCoord3f(GL.GL_TEXTURE1, t1, off, t1)
            GL.glVertex3f(RE[0], self.coord, RE[2])

            GL.glMultiTexCoord3f(GL.GL_TEXTURE0, t1, off, t0)
            GL.glMultiTexCoord3f(GL.GL_TEXTURE1, t1, off, t0)
            GL.glVertex3f(LE[0], self.coord, RE[2])

            GL.glEnd()

        GL.glUseProgram(0)

        GLUT.glutSwapBuffers()

    def move_slice(self, value):
        self.coord += value

    def rotate(self, axis, value):
        self.rotation[axis] += value/self.rfac

    def translate(self, axis, value):
        self.position[axis] += value/self.tfac
        
    def _setup_keypress_handler(self):
        self.keypress_handler = ViewHandler3D(self)
        self.keypress_handler.dispatch_table.update(dict(
            y = (self.move_slice, ( 0.05,)),
            h = (self.move_slice, (-0.05,))
            ))

if __name__ == "__main__":
    if sys.argv[-2] == '-g':
        import yt.mods
        pf = yt.mods.load(sys.argv[-1])
        main_scene = GridObject3DScene(pf)
    elif sys.argv[-2] == '-s':
        import yt.mods
        pf = yt.mods.load(sys.argv[-1])
        main_scene = GridSlice3DScene(pf)
    else:
        fn_list = glob.glob("frames/*.png")

        main_scene = MultiImageDisplayScene()
        for fn in sorted(fn_list):
            main_scene.add_image(FlatImage.from_image_file(fn))
        main_scene._current = 0
    main_scene.run()



"""
Tiger-specific fields

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields

KnownTigerFields = FieldInfoContainer()
add_tiger_field = KnownTigerFields.add_field

TigerFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = TigerFieldInfo.add_field




import numpy as np
import h5py
from yt.utilities.lib import CICSample_3
from yt.funcs import *

class ParticleGenerator(object) :

    default_fields = ["particle_position_x",
                      "particle_position_y",
                      "particle_position_z"]
    
    def __init__(self, pf, num_particles, field_list) :
        """
        Base class for generating particle fields which may be applied to
        streams. Normally this would not be called directly, since it doesn't
        really do anything except allocate memory. Takes a *pf* to serve as the
        basis for determining grids, the number of particles *num_particles*,
        and a list of fields, *field_list*.
        """
        self.pf = pf
        self.num_particles = num_particles
        self.field_list = field_list
        self.field_list.append("particle_index")
        
        try :
            self.posx_index = self.field_list.index(self.default_fields[0])
            self.posy_index = self.field_list.index(self.default_fields[1])
            self.posz_index = self.field_list.index(self.default_fields[2])
        except :
            raise KeyError("Field list must contain the following fields: " +
                           "\'particle_position_x\', \'particle_position_y\'" +
                           ", \'particle_position_z\' ")
        self.index_index = self.field_list.index("particle_index")
        
        self.num_grids = self.pf.h.num_grids
        self.NumberOfParticles = np.zeros((self.num_grids), dtype='int64')
        self.ParticleGridIndices = np.zeros(self.num_grids + 1, dtype='int64')
        
        self.num_fields = len(self.field_list)
        
        self.particles = np.zeros((self.num_particles, self.num_fields),
                                  dtype='float64')

    def has_key(self, key) :
        """
        Check to see if *key* is in the particle field list.
        """
        return (key in self.field_list)
            
    def keys(self) :
        """
        Return the list of particle fields.
        """
        return self.field_list
    
    def __getitem__(self, key) :
        """
        Get the field associated with key.
        """
        return self.particles[:,self.field_list.index(key)]
    
    def __setitem__(self, key, val) :
        """
        Sets a field to be some other value. Note that we assume
        that the particles have been sorted by grid already, so
        make sure the setting of the field is consistent with this.
        """
        self.particles[:,self.field_list.index(key)] = val[:]
                
    def __len__(self) :
        """
        The number of particles
        """
        return self.num_particles

    def get_for_grid(self, grid) :
        """
        Return a dict containing all of the particle fields in the specified grid.
        """
        ind = grid.id-grid._id_offset
        start = self.ParticleGridIndices[ind]
        end = self.ParticleGridIndices[ind+1]
        return dict([(field, self.particles[start:end,self.field_list.index(field)])
                     for field in self.field_list])
    
    def _setup_particles(self,x,y,z,setup_fields=None) :
        """
        Assigns grids to particles and sets up particle positions. *setup_fields* is
        a dict of fields other than the particle positions to set up. 
        """
        particle_grids, particle_grid_inds = self.pf.h.find_points(x,y,z)
        idxs = np.argsort(particle_grid_inds)
        self.particles[:,self.posx_index] = x[idxs]
        self.particles[:,self.posy_index] = y[idxs]
        self.particles[:,self.posz_index] = z[idxs]
        self.NumberOfParticles = np.bincount(particle_grid_inds,
                                             minlength=self.num_grids)
        if self.num_grids > 1 :
            np.add.accumulate(self.NumberOfParticles.squeeze(),
                              out=self.ParticleGridIndices[1:])
        else :
            self.ParticleGridIndices[1] = self.NumberOfParticles.squeeze()
        if setup_fields is not None:
            for key, value in setup_fields.items():
                if key not in self.default_fields:
                    self.particles[:,self.field_list.index(key)] = value[idxs]
    
    def assign_indices(self, function=None, **kwargs) :
        """
        Assign unique indices to the particles. The default is to just use
        numpy.arange, but any function may be supplied with keyword arguments.
        """
        if function is None :
            self.particles[:,self.index_index] = np.arange((self.num_particles))
        else :
            self.particles[:,self.index_index] = function(**kwargs)
            
    def map_grid_fields_to_particles(self, mapping_dict) :
        r"""
        For the fields in  *mapping_dict*, map grid fields to the particles
        using CIC sampling.

        Examples
        --------
        >>> field_map = {'Density':'particle_density',
        >>>              'Temperature':'particle_temperature'}
        >>> particles.map_grid_fields_to_particles(field_map)
        """
        pbar = get_pbar("Mapping fields to particles", self.num_grids)
        for i, grid in enumerate(self.pf.h.grids) :
            pbar.update(i)
            if self.NumberOfParticles[i] > 0:
                start = self.ParticleGridIndices[i]
                end = self.ParticleGridIndices[i+1]
                # Note we add one ghost zone to the grid!
                cube = grid.retrieve_ghost_zones(1, mapping_dict.keys())
                le = np.array(grid.LeftEdge).astype(np.float64)
                dims = np.array(grid.ActiveDimensions).astype(np.int32)
                for gfield, pfield in mapping_dict.items() :
                    field_index = self.field_list.index(pfield)
                    CICSample_3(self.particles[start:end,self.posx_index],
                                self.particles[start:end,self.posy_index],
                                self.particles[start:end,self.posz_index],
                                self.particles[start:end,field_index],
                                np.int64(self.NumberOfParticles[i]),
                                cube[gfield], le, dims,
                                np.float64(grid['dx']))
        pbar.finish()

    def apply_to_stream(self, clobber=False) :
        """
        Apply the particles to a stream parameter file. If particles already exist,
        and clobber=False, do not overwrite them, but add the new ones to them. 
        """
        grid_data = []
        for i,g in enumerate(self.pf.h.grids) :
            data = {}
            if clobber :
                data["number_of_particles"] = self.NumberOfParticles[i]
            else :
                data["number_of_particles"] = self.NumberOfParticles[i] + \
                                              g.NumberOfParticles
            grid_particles = self.get_for_grid(g)
            for field in self.field_list :
                if data["number_of_particles"] > 0 :
                    # We have particles in this grid
                    if g.NumberOfParticles > 0 and not clobber:
                        # Particles already exist
                        if field in self.pf.h.field_list :
                            # This field already exists
                            prev_particles = g[field]
                        else :
                            # This one doesn't, set the previous particles' field
                            # values to zero
                            prev_particles = np.zeros((g.NumberOfParticles))
                        data[field] = np.concatenate((prev_particles,
                                                      grid_particles[field]))
                    else :
                        # Particles do not already exist or we're clobbering
                        data[field] = grid_particles[field]
                else :
                    # We don't have particles in this grid
                    data[field] = np.array([], dtype='float64')
            grid_data.append(data)
        self.pf.h.update_data(grid_data)

class FromListParticleGenerator(ParticleGenerator) :

    def __init__(self, pf, num_particles, data) :
        r"""
        Generate particle fields from array-like lists contained in a dict.

        Parameters
        ----------
        pf : `StaticOutput`
            The parameter file which will serve as the base for these particles.
        num_particles : int
            The number of particles in the dict.
        data : dict of NumPy arrays
            The particle fields themselves.

        Examples
        --------
        >>> num_p = 100000
        >>> posx = np.random.random((num_p))
        >>> posy = np.random.random((num_p))
        >>> posz = np.random.random((num_p))
        >>> mass = np.ones((num_p))
        >>> data = {'particle_position_x': posx, 'particle_position_y': posy,
        >>>         'particle_position_z': posz, 'particle_mass': mass}
        >>> particles = FromListParticleGenerator(pf, num_p, data)
        """

        field_list = data.keys()
        x = data.pop("particle_position_x")
        y = data.pop("particle_position_y")
        z = data.pop("particle_position_z")

        xcond = np.logical_or(x < pf.domain_left_edge[0],
                              x >= pf.domain_right_edge[0])
        ycond = np.logical_or(y < pf.domain_left_edge[1],
                              y >= pf.domain_right_edge[1])
        zcond = np.logical_or(z < pf.domain_left_edge[2],
                              z >= pf.domain_right_edge[2])
        cond = np.logical_or(xcond, ycond)
        cond = np.logical_or(zcond, cond)

        if np.any(cond) :
            raise ValueError("Some particles are outside of the domain!!!")

        ParticleGenerator.__init__(self, pf, num_particles, field_list)
        self._setup_particles(x,y,z,setup_fields=data)
        
class LatticeParticleGenerator(ParticleGenerator) :

    def __init__(self, pf, particles_dims, particles_left_edge,
                 particles_right_edge, field_list) :
        r"""
        Generate particles in a lattice arrangement. 

        Parameters
        ----------
        pf : `StaticOutput`
            The parameter file which will serve as the base for these particles.
        particles_dims : int, array-like 
            The number of particles along each dimension
        particles_left_edge : float, array-like
            The 'left-most' starting positions of the lattice.
        particles_right_edge : float, array-like
             The 'right-most' ending positions of the lattice.
        field_list : list of strings
             A list of particle fields
             
        Examples
        --------
        >>> dims = (128,128,128)
        >>> le = np.array([0.25,0.25,0.25])
        >>> re = np.array([0.75,0.75,0.75])
        >>> fields = ["particle_position_x","particle_position_y",
        >>>           "particle_position_z",
        >>>           "particle_density","particle_temperature"]
        >>> particles = LatticeParticleGenerator(pf, dims, le, re, fields)
        """

        num_x = particles_dims[0]
        num_y = particles_dims[1]
        num_z = particles_dims[2]
        xmin = particles_left_edge[0]
        ymin = particles_left_edge[1]
        zmin = particles_left_edge[2]
        xmax = particles_right_edge[0]
        ymax = particles_right_edge[1]
        zmax = particles_right_edge[2]

        xcond = (xmin < pf.domain_left_edge[0]) or \
                (xmax >= pf.domain_right_edge[0])
        ycond = (ymin < pf.domain_left_edge[1]) or \
                (ymax >= pf.domain_right_edge[1])
        zcond = (zmin < pf.domain_left_edge[2]) or \
                (zmax >= pf.domain_right_edge[2])
        cond = xcond or ycond or zcond

        if cond :
            raise ValueError("Proposed bounds for particles are outside domain!!!")

        ParticleGenerator.__init__(self, pf, num_x*num_y*num_z, field_list)

        dx = (xmax-xmin)/(num_x-1)
        dy = (ymax-ymin)/(num_y-1)
        dz = (zmax-zmin)/(num_z-1)
        inds = np.indices((num_x,num_y,num_z))
        xpos = inds[0]*dx + xmin
        ypos = inds[1]*dy + ymin
        zpos = inds[2]*dz + zmin
        
        self._setup_particles(xpos.flat[:], ypos.flat[:], zpos.flat[:])
        
class WithDensityParticleGenerator(ParticleGenerator) :

    def __init__(self, pf, data_source, num_particles, field_list,
                 density_field="Density") :
        r"""
        Generate particles based on a density field.

        Parameters
        ----------
        pf : `StaticOutput`
            The parameter file which will serve as the base for these particles.
        data_source : `yt.data_objects.api.AMRData`
            The data source containing the density field.
        num_particles : int
            The number of particles to be generated
        field_list : list of strings
            A list of particle fields
        density_field : string, optional
            A density field which will serve as the distribution function for the
            particle positions. Theoretically, this could be any 'per-volume' field. 
            
        Examples
        --------
        >>> sphere = pf.h.sphere(pf.domain_center, 0.5)
        >>> num_p = 100000
        >>> fields = ["particle_position_x","particle_position_y",
        >>>           "particle_position_z",
        >>>           "particle_density","particle_temperature"]
        >>> particles = WithDensityParticleGenerator(pf, sphere, num_particles,
        >>>                                          fields, density_field='Dark_Matter_Density')
        """

        ParticleGenerator.__init__(self, pf, num_particles, field_list)

        num_cells = len(data_source["x"].flat)
        max_mass = (data_source[density_field]*
                    data_source["CellVolume"]).max()
        num_particles_left = num_particles
        all_x = []
        all_y = []
        all_z = []
        
        pbar = get_pbar("Generating Particles", num_particles)
        tot_num_accepted = int(0)
        
        while num_particles_left > 0:

            m = np.random.uniform(high=1.01*max_mass,
                                  size=num_particles_left)
            idxs = np.random.random_integers(low=0, high=num_cells-1,
                                             size=num_particles_left)
            m_true = (data_source[density_field]*
                      data_source["CellVolume"]).flat[idxs]
            accept = m <= m_true
            num_accepted = accept.sum()
            accepted_idxs = idxs[accept]
            
            xpos = data_source["x"].flat[accepted_idxs] + \
                   np.random.uniform(low=-0.5, high=0.5, size=num_accepted) * \
                   data_source["dx"].flat[accepted_idxs]
            ypos = data_source["y"].flat[accepted_idxs] + \
                   np.random.uniform(low=-0.5, high=0.5, size=num_accepted) * \
                   data_source["dy"].flat[accepted_idxs]
            zpos = data_source["z"].flat[accepted_idxs] + \
                   np.random.uniform(low=-0.5, high=0.5, size=num_accepted) * \
                   data_source["dz"].flat[accepted_idxs]

            all_x.append(xpos)
            all_y.append(ypos)
            all_z.append(zpos)

            num_particles_left -= num_accepted
            tot_num_accepted += num_accepted
            pbar.update(tot_num_accepted)

        pbar.finish()

        x = np.concatenate(all_x)
        y = np.concatenate(all_y)
        z = np.concatenate(all_z)

        self._setup_particles(x,y,z)
        



"""
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation:  UCSD
License:
  Copyright (C) 2010-2011 Matthew Turk  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import OpenGL.GL as GL
import OpenGL.GLUT as GLUT
import OpenGL.GLU as GLU
import OpenGL.GL.shaders as shaders
from OpenGL.arrays import vbo, ArrayDatatype
import OpenGL.GL.ARB.framebuffer_object as GL_fbo
from functools import wraps

def render_fbo(func):
    @wraps(func)
    def fbo_renderer(self, fbo_id = "fbo_id", fbo_depth = "fbo_depth", **kwargs):
        GL_fbo.glBindFramebuffer(GL_fbo.GL_FRAMEBUFFER,
                                 self.gl_state[fbo_id])
        GL_fbo.glBindRenderbuffer(GL_fbo.GL_RENDERBUFFER,
                                  self.gl_state[fbo_depth])
        self.set_viewport()
        GL.glPushAttrib(GL.GL_VIEWPORT_BIT)
        GL.glViewport(0, 0, self.gl_state["width"], self.gl_state["height"])
        status = GL_fbo.glCheckFramebufferStatus(GL_fbo.GL_FRAMEBUFFER)
        assert(status == GL_fbo.GL_FRAMEBUFFER_COMPLETE)

        func(self, **kwargs)

        GL.glPopAttrib(GL.GL_VIEWPORT_BIT)
        GL_fbo.glBindFramebuffer(GL_fbo.GL_FRAMEBUFFER, 0)

    return fbo_renderer

def create_fbo(gl_state, fbo_id = "fbo_id", fbo_depth = "fbo_depth",
               fbo_tex = "fbo_tex"):
    GL.glActiveTexture(GL.GL_TEXTURE0)
    id_fbo = GL.glGenTextures(1)
    id_depth = GL_fbo.glGenRenderbuffers(1)
    gl_state[fbo_tex] = id_fbo
    gl_state[fbo_depth] = id_depth
    gl_state[fbo_id] = GL_fbo.glGenFramebuffers(1)

    GL.glBindTexture(GL.GL_TEXTURE_2D, id_fbo)
    GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
    GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE)
    GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE)
    GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
    GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
    GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA8,
                    gl_state["width"], gl_state["height"], 0,
                    GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, None)

    GL_fbo.glBindFramebuffer(GL_fbo.GL_FRAMEBUFFER, gl_state[fbo_id])
    GL_fbo.glFramebufferTexture2D(GL_fbo.GL_FRAMEBUFFER,
        GL_fbo.GL_COLOR_ATTACHMENT0, GL.GL_TEXTURE_2D, gl_state[fbo_tex], 0)

    GL_fbo.glBindRenderbuffer(GL_fbo.GL_RENDERBUFFER, id_depth)
    GL_fbo.glRenderbufferStorage(GL_fbo.GL_RENDERBUFFER,
        GL.GL_DEPTH_COMPONENT, gl_state["width"], gl_state["height"])
    GL_fbo.glFramebufferRenderbuffer(
        GL_fbo.GL_FRAMEBUFFER, GL_fbo.GL_DEPTH_ATTACHMENT,
        GL_fbo.GL_RENDERBUFFER, id_depth)

    GL_fbo.glBindFramebuffer(GL_fbo.GL_FRAMEBUFFER, 0)
    GL_fbo.glBindRenderbuffer(GL_fbo.GL_RENDERBUFFER, 0)

def identity_view(func):
    pass

def translate_view(func):
    pass
